#!/etc/ConsolePi/venv/bin/python3

import re
import sys
from os import system
from typing import Dict, Iterator, List, Tuple, Union, Any

from consolepi import utils, log, config  # type: ignore

MIN_WIDTH = 55
MAX_COLS = 5
DEF_LEGEND_OPTIONS = {
    "back": ("b", "Back"),
    "next": ("n", "Next Page"),
    # "tl": ("tl", "Toggle (show/hide) Legend"),
    "x": ("x", "Exit")
}
COL_PAD = 3
L_OFFSET = 1  # Number of spaces the body of menu is offset from left.
MIN_LINES_FOR_SPLIT_COL = 2  # Min number of item entries that need to be avail in col to split next group


class Line:
    """Constructor Class for line object"""

    def __init__(self, line_len: int, line_text: str) -> None:
        self.len = line_len
        self.text = line_text


def format_line(line: Union[str, bool]) -> Line:
    """Format line for display in menu.

    Args:
        line (str|bool): If str is provided supports formatting placeholders {{red}}, {{blink}}, etc.
                            If bool is provided returns colorized [ON] (green), or [OFF] (red) string.

    Returns:
        object: Line object with len and text attributes, where len is the effective printed len with
                any ASCII formatting characters stripped, and line is formatted (with ASCII colors)
    """

    if isinstance(line, bool):
        line = "{{green}}ON{{norm}}" if line else "{{red}}OFF{{norm}}"

    colors = {
        "green": "\033[1;32m",  # Bold with normal ForeGround
        "red": "\033[1;31m",
        "yellow": "\033[1;33m",
        "blue": "\033[1;34m",
        "magenta": "\033[1;35m",
        "cyan": "\033[1;36m",
        "dgreen": "\033[2;32m",  # Dim with normal ForeGround
        "dred": "\033[2;31m",
        "dyellow": "\033[2;33m",
        "dblue": "\033[2;34m",
        "dmagenta": "\033[2;35m",
        "dcyan": "\033[2;36m",
        "inverted": "\033[7m",
        "norm": "\033[0m",  # Reset to Normal
        "dot": "\u00B7",  # no it's not a color it's a dot
    }

    _l = line
    if "{{" in line:
        for c in colors:
            _l = _l.replace(
                "{{" + c + "}}", ""
            )  # color format var removed to calc effective line length
            line = line.replace(
                "{{" + c + "}}", colors[c]
            )  # line formatted with coloring

    return Line(len(_l), line)


class TTY:
    def __init__(self) -> None:
        self.rows: Union[int, None] = None
        self.cols: Union[int, None] = None
        self.update()

    def __str__(self):
        return "\n".join([
            '',
            'TTY Size:',
            f'  rows: {self.rows}',
            f'  cols: {self.cols}',
            ''
        ])

    def update(self) -> None:
        if sys.stdin.isatty():
            self.rows, self.cols = utils.get_tty_size()
        else:
            self.rows = self.cols = None

    @staticmethod
    def __bool__() -> bool:
        return sys.stdin.isatty()


tty = TTY()


#
#  A section of the menu
#  Header, subhead, body, legend, footer
#
class MenuSection:
    def __init__(self, orig: Any = None, lines: list = [], width_list: list = [], rows: int = 0, cols: int = 0,
                 opts: list = [], overrides: dict = {}, update_method: Any = None, update_args: Union[list, tuple, set] = [],
                 update_kwargs: dict = {}, name: str = None, hide: bool = False) -> None:
        self.hide = hide if name != "legend" else hide or config.hide_legend
        self.orig = orig  # Original unformatted data
        self.lines = lines
        self.width_list = width_list
        self.opts = opts
        self.overrides = overrides
        self.rows = rows
        self.cols = cols
        self.name_in = name
        self.update_method = update_method
        self.update_args = update_args
        self.update_kwargs = update_kwargs

    def __repr__(self):
        if self.name_in:
            name = f" ({self.name_in})"
        else:
            name = "" if not self.update_method else f" ({repr(self.update_method).split('format_')[-1].split()[0]})"
        return f"<{self.__module__}.{type(self).__name__}{name} object at {hex(id(self))}>"

    def __str__(self):
        return "\n".join(self.lines)

    def __len__(self):
        return len(self.lines)

    def __bool__(self):
        return bool(self.lines)

    def __add__(self, data: Union[int, list]):
        if isinstance(data, list):
            line_info = [format_line(line) for line in data]
            self.width_list += [line.len for line in line_info]
            self.lines += [line.text for line in line_info]
            self.update()

    def __getitem__(self, index: Union[int, list, slice, tuple, set]) -> Any:
        if isinstance(index, (int, slice)):
            ret = self.lines[index]
        else:
            try:
                ret = [self.lines[i] for i in index]
            except Exception as e:
                ret = e
        return ret

    @property
    def name(self):
        return self.name_in if self.name_in else self._name()

    def _name(self):
        name = repr(self) if not self.update_method else f"{repr(self.update_method).split('format_')[-1].split()[0]}"
        return name

    def update(self, *args, **kwargs):
        if self.update_method:
            self.update_method(*[*self.update_args, *args], **{**self.update_kwargs, **kwargs})

    def append(self, data: str):
        line = format_line(data)
        self.lines += [line.text]
        self.width_list += [line.len]
        if line.len > self.cols:
            self.cols = line.len
        self.update()

    def get(self, attr: str, default: Any = None):
        if hasattr(self, attr):
            return getattr(self, attr)
        # elif attr in self.data_dict:
        #     return self.data_dict[attr]
        elif default:
            return default
        else:
            raise AttributeError(f'{self.__class__.__name__} does not have attribute {attr}')


#
#   The Primary Menu Object
#   Represents page
#
class MenuParts:
    def __init__(self, name: str = None):
        self.name = name
        self._header = None
        self._subhead = None
        self._body = None
        self._legend = None
        self._footer = None
        self.prev_slice = {}
        self.this_slice = {}
        self.next_slice = {}
        self.parts: Union[List[MenuSection], List[None]] = [
            self.header,
            self.subhead,
            self.body,
            self.legend,
            self.footer
        ]
        # +1 is for prompt line
        self.rows = 0 if len(self) == 0 else (len(self) + 1)
        self.cols = self._cols()
        # tty = tty

    @property
    def header(self):
        return self._header

    @header.setter
    def header(self, new_value):
        self._header = new_value
        self.update()
        # self._header.update(width=self.cols)

    @property
    def subhead(self):
        return self._subhead

    @subhead.setter
    def subhead(self, new_value):
        self._subhead = new_value
        self.update()
        # self._subhead.update(width=self.cols)

    @property
    def body(self):
        return self._body

    @body.setter
    def body(self, new_value):
        self._body = new_value
        self.update()
        # self._body.update(width=self.cols)

    @property
    def legend(self):
        return self._legend

    @legend.setter
    def legend(self, new_value):
        self._legend = new_value
        self.update()
        # self._legend.update(width=self.cols)

    @property
    def footer(self):
        return self._footer

    @footer.setter
    def footer(self, new_value):
        self._footer = new_value
        self.update()
        # self._footer.update(width=self.cols)

    def __repr__(self):
        name = "" if not self.name else f" ({(self.name)})"
        return f"<{self.__module__}.{type(self).__name__}{name} object at {hex(id(self))}>"

    def __len__(self):
        # parts = [self.header, self.subhead, self.body, self.legend, self.footer]
        return sum([p.rows for p in self.parts if p] or [0])

    def __str__(self):
        # ---- CLEAR SCREEN -----
        if not config.debug:
            _ = system("clear -x")
        else:
            print("")  # if DEBUG need this to get off the prompt line

        parts = [self.header, self.subhead, self.body, self.legend, self.footer]
        if self.legend and self.legend.hide:
            _ = parts.pop(parts.index(self.legend))
        log.clear()
        return "\n".join([line for p in parts for line in p.lines])

    def __iter__(self, key: str = None) -> Iterator[MenuSection]:
        # parts = [self.header, self.subhead, self.body, self.legend, self.footer]
        for p in self.parts:
            if not p or (key and p.name != key):  # or (p.name == "legend" and p.hide):
                continue
            else:
                yield p

    def _cols(self):
        # parts = [self.header, self.subhead, self.body, self.legend, self.footer]
        _ = "Do Nothing Break Point Line"  # TODO remove
        return max([p.cols for p in self.parts if p] or [0])

    def update(self):
        self.rows = 0 if len(self) == 0 else (len(self) + 1)
        self.cols = self._cols()
        if self.legend and self.legend.hide:
            if self.legend in self.parts:
                _ = self.parts.pop(self.parts.index(self.legend))
            parts = [self.header, self.subhead, self.footer]
        else:
            self.parts = [self.header, self.subhead, self.body, self.legend, self.footer]
            parts = [self.header, self.subhead, self.legend, self.footer]

        self.body_avail_rows = tty.rows - sum([p.rows for p in parts if p] or [0]) - 1  # -1 for prompt line

    def diag(self, diag_data: List[str] = None):
        tty.update()
        parts = [self.header, self.subhead, self.body, self.legend, self.footer]
        ret = ["", ""]
        if diag_data:
            ret += diag_data

        ret += [f"    tty rows, cols: {tty.rows} {tty.cols}"]
        ret += [
            f"    {p.name} rows, cols: {p.rows} {p.cols}"
            f"{'' if not p.name == 'body' else f' (Available Rows: {self.body_avail_rows})'}"
            for p in parts
        ]
        ret += [f"    Total Menu rows, cols: {self.rows} {self.cols}", ""]
        return "\n".join(ret)


class Menu:
    def __init__(self, name: str = None, left_offset: int = L_OFFSET):  # utils, debug=False, log=None, log_file=None):
        self.name = name
        self.left_offset = left_offset
        self.go: bool = True
        self.def_actions = {
            "dump": self.dump_formatter_data,
            "tl": self.toggle_legend
        }
        self.states = {True: "{{green}}ON{{norm}}", False: "{{red}}OFF{{norm}}"}
        self.ignored_errors: List[Union[str, re.Pattern]] = []  # Populated by menu script consolepi-menu.py
        self.log_sym_2bang: str = "\033[1;33m!!\033[0m"
        self.prev_page: int = 1
        self.cur_page: int = 1
        self.legend_options: Union[Dict[str, List[str]], None] = None
        self.init_pager()
        self.page = MenuParts(name)  # Needs to come after init_pager()
        self.actions: Union[Dict[str, str], None] = None  # set in print_menu()
        self.body_in = None
        self.items_in = None
        self.subs_in = None
        self.legend_in = None
        self.menu_actions_in: Union[Dict[str, str], None] = None
        self.pbody = None
        self.pitems = None
        self.psubs: Union[List[str], None] = None
        self.reverse = False
        self.prev_header = None
        self.tot_body_1col_rows = 0
        self.pages = {}
        self.item = 1
        self.tty = tty

    def __repr__(self):
        name = "" if not self.name else f" ({self.name})"
        return f"<{self.__module__}.{type(self).__name__}{name} object at {hex(id(self))}>"

    def init_pager(self):
        self.prev_page = self.cur_page
        self.prev_col_width: int = 0  # assigned in do_pager
        self.col_width: int = 0       # assigned in do_pager
        self.page_width: int = 0      # assigned in do_pager
        self.cur_page_width: int = 0  # assigned in pager_write_other_col

    def print_menu(
        self,
        body: list,
        subs: Union[List[str], None] = None,
        header: Union[str, list] = None,
        subhead: Union[str, list] = None,
        legend: Union[dict, list] = None,
        format_subs: bool = False,
        by_tens: bool = False,
        menu_actions: dict = {},
        hide_legend: bool = None,
    ) -> dict:
        """Format and print current menu, sized to fit terminal.  Pager implemented if required.

        Args:
            body (list): A list of lists or list of strings, where each inner list is made up of text for each menu-item
                         in that logical section/group.
            subs (list, optional): A list of sub-head lines that map to each inner body list.  This is the header for the
                                   specific logical grouping of menu-items. Should be of equal len to body list.
                                   Defaults to None.
            header (str|list, optional): The main header text for the menu. Multi-Line header via list of strings.
                                         Defaults to None.
            subhead (str|list, optional): Subhead text is printed below the main menu header with blank lines added before and
                                       after. Defaults to None.
            legend (dict|list, optional): If a list is provided it should be a list of strings matching keys in the class
                                          attribute legend_options.
                                          If dict is provided legend = {
                                              "opts": [keys], "overrides": {
                                                  "key": ("item", "description")
                                                  }
                                              }
                                          Where item and description are what prints in the legend.
                                          Defaults to None.
            col_pad (int, optional): The amount of spacing inserted in betweeen each vertical column in the menu. The col_pad is
                                     reduced to as low as 1 if menu overruns tty size.
                                     Defaults to 4.
            format_subs (bool, optional): Only Applies to sub-head auto formatting.
                                        Auto formatting results in '------- sub text -------' (width of section)
                                        Defaults to True.
            by_tens (bool, optional): Will start each section @ 1, 11, 21, 31... unless the section is greater than 10.
                                      Defaults to False.
            menu_actions (dict, optional): The Actions dict (TODO make object).  Determines what function is called when a menu
                                           item is selected Defaults to {}.
            menu_actions (bool, optional): Override config option for menus where you always want legend printed.

        Returns:
            [dict]: Returns the menu_actions dict which may include paging actions if warranted.
        """
        self.tty.update()

        do_size = False
        if not self.body_in:
            self.body_in = body
            do_size = True

        refresh = False
        if not self.subs_in:
            self.subs_in = subs

        # if a refresh triggered a change to subs/body reset to page 1
        elif subs:
            if len(body) != len(self.body_in):
                refresh = True
                self.subs_in = subs
                self.body_in = body
                self.reverse = False
                self.pbody, self.psubs, self.pitems = None, None, None
                self.prev_page, self.cur_page = 1, 1
                self.page.prev_slice = {}
                self.reverse = False

        # This sets the menu items (the numbers the user selects) as a list of List[int]
        if not self.items_in or refresh:
            x = 1
            self.items_in = []
            for idx, sec in enumerate(self.body_in):
                if idx > 0:
                    x += len(self.body_in[idx - 1])
                self.items_in.append([y + x for y in range(len(sec))])

        if menu_actions:
            self.menu_actions_in = menu_actions
            self.actions = None

        self.actions = self.menu_actions_in if not self.actions else {**self.menu_actions_in, **self.actions}

        if not self.legend_in or self.legend_in != legend:
            if isinstance(legend, dict):
                self.legend_in = legend
            elif isinstance(legend, list):
                self.legend_in = {"opts": legend}

        self.legend_options = DEF_LEGEND_OPTIONS if not self.legend_options else {**DEF_LEGEND_OPTIONS, **self.legend_options}
        header = self.format_header(header)
        subhead = self.format_subhead(subhead)
        self.format_legend(self.legend_in)
        if hide_legend is not None:
            self.page.legend.hide = hide_legend
        self.format_footer()

        if not subs:
            header.append("")

        if do_size:
            self.size = self.calc_size(self.body_in, subs, format_subs=format_subs, by_tens=by_tens)

        self.pg_cnt = 0
        try:
            self.size.get_cols(self.page.body_avail_rows)
            if len(self.size.pages) > 1:
                self.page.body_avail_rows -= 1
            self.pg_cnt = len(self.size.pages)
            self.addl_rows = self.size.addl_rows
        except Exception as e:
            log.error(f"Exception in size.get_cols() {e}")

        # if str was passed place in list to iterate over
        if isinstance(body, str):
            body = [body]

        # ensure body is a list of lists for mapping with list of subs
        _body = [body] if len(body) >= 1 and isinstance(body[0], str) else body

        # Override body and sub lists when navigating paged output
        _body = _body if not self.pbody else self.pbody
        subs = subs if not self.psubs else self.psubs
        items = self.items_in if not self.pitems else self.pitems

        if subs and len(_body) != len(subs):
            raise ValueError("body list and sub list not of equal size")

        prev_pages = self.pages
        if self.pages:
            self.pages = {k: [] for k in self.pages.keys() if k <= self.cur_page}

        # -- // EMPTY MENU No Adapters / HOSTS / REMOTES \\ --
        if not body:
            return self.empty_menu()

        # # -- // SET STARTING ITEM # \\ --
        if not self.reverse:
            item = min([item for sublist in items for item in sublist])
        else:
            item = max([item for sublist in items for item in sublist])

        if not self.reverse:
            self.page.next_slice = {}
        self.page.this_slice = {}

        start = 1 if not self.reverse else 10
        col_lines = []
        addl_rows = 0
        section_slices = {}

        equal_sections = True if by_tens or len(set([len(section) for section in _body])) == 1 else False
        max_section = max([len(section) + addl_rows for section in _body])
        for idx, _section in enumerate(_body):
            if subs:
                try:
                    sec = self.subs_in.index(subs[idx])
                except ValueError as e:
                    log.debug(f"{e.__class__.__name__}: {e}", show=log.DEBUG or config.debug)
                    match = [i for i, x in enumerate(self.subs_in) if x.startswith(subs[idx].split(" @")[0])]
                    if match and len(match) == 1:
                        sec = match[0]
                        self.subs_in[sec] = subs[idx]  # update original all subs list to include updated sub entry
                    else:
                        sec = idx
                sub = subs[idx]
            else:
                sec = idx
                sub = None

            if by_tens and _body.index(_section) > 0:
                if not self.reverse:
                    item = start + 10 if item <= start + 10 else item
                    start += 10
                else:  # TODO This is not tested
                    item = start - 10 if item >= start - 10 else item
                    start -= 10

            # -- Get # of addl_rows added to section by formatter (ran for 1st section only)
            if _body.index(_section) == 0:
                if not hasattr(self, "addl_rows"):
                    this_lines, this_width, this_rows = self.format_section(
                        body=_section,
                        sub=sub,
                        index=item,
                        format_sub=format_subs,
                    )

                    # determine how many addl rows were added by formatter
                    addl_rows = this_rows - len(_section)
                else:
                    addl_rows = self.addl_rows
                # determine number of rows if menu printed as 1 col
                self.tot_body_1col_rows = sum([len(_section) + addl_rows for _section in _body])

            # current col must have room for 3 items from section otherwise new section starts in next col
            if self.page.body_avail_rows - len(col_lines) >= 3 + addl_rows:
                if (len(col_lines) + len(_section) + addl_rows) >= self.tot_body_1col_rows / 2 and \
                        (len(col_lines) + len(_section) + addl_rows) <= self.page.body_avail_rows:
                    _end = len(_section)
                elif (len(col_lines) + len(_section) + addl_rows) >= self.page.body_avail_rows:
                    if self.pg_cnt == 1 and len(_body) <= 3 and max_section <= self.page.body_avail_rows:
                        _end = len(_section)
                    else:                                       # HACK the -1 below is a hack
                        _end = self.page.body_avail_rows - (len(col_lines) + addl_rows) - 1  # FIXME avail rows 52 cols 190 hide legend.  wsl menu 108 through sec 15 on 1st page 2nd page sec 16 gets stripped to slice(12, 13, 1) rather than (0, 13, 1)
                elif len(_section) + addl_rows <= self.page.body_avail_rows:             # It's the first col of the next page shouldn't be stripped.  _end should be 11 it's being set to 12 body_avail_rows is 45 disregarding space at top and bot body is 43
                    _end = len(_section)
                else:
                    _end = self.page.body_avail_rows - len(col_lines) - addl_rows
            elif self.pg_cnt == 1 and len(_body) <= 3 and max_section <= self.page.body_avail_rows:
                _end = len(_section)
            else:
                _end = self.page.body_avail_rows - addl_rows

            # -- break sections with too many rows to fit tty into list of slices for each col that fit into limits of tty
            if 0 < _end < len(_section):
                _slice = [slice(0, _end, 1)]
                while len(_section[_end:len(_section)]) + addl_rows > self.page.body_avail_rows:
                    _start, _end = _end, _end + (self.page.body_avail_rows - addl_rows)
                    _slice += [slice(_start, _end, 1)]

                # Add any remaining slices in section
                if _end < len(_section):
                    _slice += [slice(_end, len(_section), 1)]
            else:
                _slice = [slice(0, len(_section), 1)]

            for s in _slice:
                sub_section = _section[s]

                # Update sub-header for section if CONTINUED
                _sub_key = 0 if not self.reverse else -1
                try:
                    if sub and sub_section[_sub_key].split()[0] != self.body_in[sec][0].split()[0]:
                        if "Local Adapters" in sub:
                            # if self.name != "rename_menu":
                            sub = f"[CONTINUED] {sub.replace('Rename ', '')}"
                        else:
                            sub = f"[CONTINUED] {sub.split('] ')[-1].split(' @')[0].split(' on ')[-1]}"
                except Exception as e:
                    sub = f"{sub} <-- Exception ({e.__class__.__name__})"

                # if debug enabled Add index of section to section header (sub)
                if sub and config.debug and not sub.startswith(f"-{sec}-"):
                    sub = f"-{sec}-{sub}"

                this_lines, this_width, this_rows = self.format_section(
                    body=sub_section,
                    sub=sub,
                    index=item,
                    format_sub=format_subs
                )

                item = item + len(sub_section) if not self.reverse else item - len(sub_section)

                # Broken out this way for easier debug / verification of logic (step through in debugger)
                if col_lines and len(col_lines) + this_rows > self.page.body_avail_rows:
                    self.pager_write_col_to_page(col_lines, section_slices)
                    section_slices = {}
                    # -- Prev Col written update col with current lines
                    col_lines = this_lines
                    self.col_width = this_width
                elif col_lines and len(col_lines) + len(this_lines) >= self.tot_body_1col_rows / 2 and (
                    not self.pages or (
                        len(self.pages[self.cur_page]) > 0 and not (
                            len(col_lines) + len(_section) + addl_rows) <= len(self.pages[self.cur_page])
                    )
                ):
                    self.pager_write_col_to_page(col_lines, section_slices)
                    section_slices = {}
                    # -- Prev Col written update col with current lines
                    col_lines = this_lines
                    self.col_width = this_width
                elif col_lines and max_section <= self.page.body_avail_rows and equal_sections:
                    self.pager_write_col_to_page(col_lines, section_slices)
                    section_slices = {}
                    # -- Prev Col written update col with current lines
                    col_lines = this_lines
                    self.col_width = this_width
                elif col_lines and sum([len(s) + addl_rows for s in _body[idx:]]) <= len(col_lines) and \
                        not len(col_lines) + len(this_lines) <= len(self.pages[self.cur_page]):
                    self.pager_write_col_to_page(col_lines, section_slices)
                    section_slices = {}
                    # -- Prev Col written update col with current lines
                    col_lines = this_lines
                    self.col_width = this_width
                elif col_lines and self.cur_page == 1 and self.pg_cnt == 1 and \
                        len(_body) <= 3 and max_section <= self.page.body_avail_rows:
                    self.pager_write_col_to_page(col_lines, section_slices)
                    section_slices = {}
                    # -- Prev Col written update col with current lines
                    col_lines = this_lines
                    self.col_width = this_width
                else:  # -- Appending to Existing Column Col not written to Page Yet
                    if not self.reverse:
                        col_lines += this_lines
                    else:
                        col_lines = this_lines + col_lines

                    if this_width > self.col_width:
                        self.col_width = this_width

                if self.cur_page_width >= self.page.legend.cols:
                    self.page.legend.lines = [f"{line:{self.cur_page_width}}" for line in self.page.legend.lines]
                    self.page.legend.cols = self.cur_page_width

                section_slices[sec] = s if not self.reverse else \
                    slice(len(self.body_in[sec]) - s.stop, len(self.body_in[sec]) - s.start, 1)

        # Write Final Col
        self.pager_write_col_to_page(col_lines, section_slices)

        if self.reverse:
            self.pages = {**prev_pages, **self.pages}

        self.page.body = MenuSection(
            lines=self.pages[self.cur_page],
            rows=len(self.pages[self.cur_page]),
            cols=self.cur_page_width,
            name="body"
        )

        # Adds Next/Back option as appropriate for Paged Menu
        self.pager_update_legend()

        # update slices (first/last items) in the event tty was resized and no page change was selected
        self.update_slices()

        # FIXME this was issue with con menu where legend was widest section
        self.cur_page_width = max([self.cur_page_width, self.page.legend.cols, self.page.footer.cols, self.page.header.cols])

        for menu_part in [self.page.legend, self.page.footer, self.page.header]:
            menu_part.update(width=self.cur_page_width)
            # menu_part.update(width=self.page.cols)

        # -- // PRINT THE MENU \\ --
        print(self.page)

        self.init_pager()

        return {**self.def_actions, **self.actions}

    def dump_formatter_data(self):
        diag_data = [f"    Total Rows(1 col): {self.tot_body_1col_rows}"]
        print(self.page.diag(diag_data))
        input('Press Enter to Continue... ')

    def empty_menu(self) -> None:
        '''Print Menu with Empty body when no body is provided to print_menu().
        '''
        self.page.body = MenuSection(
            lines=[],
            rows=0,
            cols=0,
            name="body"
        )

        self.format_subhead(
            [
                "",
                "No local serial devices found.",
                "No remote ConsolePis discovered and reachable.",
                "No Manually Defined TELNET/SSH hosts configured.",
                f"{self.log_sym_2bang}  There is Nothing to display  {self.log_sym_2bang}"
            ]
        )
        opts = ["x"] if not self.legend_in or "refresh" not in self.legend_in.get("opts", []) else ["refresh", "x"]
        self.page.legend.update(legend={"opts": opts}, width=self.page.cols)

        for menu_part in [self.page.legend, self.page.footer, self.page.header]:
            menu_part.update(width=self.page.cols)

        print(self.page)
        self.init_pager()

        return self.menu_actions_in

    def calc_size(self, body: list, subs: Union[list, None],
                  format_subs: bool = False, by_tens: bool = False) -> object:
        # tty = tty
        body = utils.listify(body)
        body = [body] if len(body) >= 1 and isinstance(body[0], str) else body
        reverse = self.reverse
        format_section = self.format_section
        cur_page = self.cur_page
        body_avail_rows = self.page.body_avail_rows
        left_offset = self.left_offset

        class Size:
            def __init__(self):
                self.body_avail_rows = body_avail_rows
                self.left_offset = left_offset
                self.lines = []
                self.cols = []
                self.rows = []
                self.pages = {}
                self.col_width: int = 0
                self.page_width: int = 0
                # self.vert_cols: int = 0  # TODO Remove once verified no side effects
                self.page: int = 1
                self.body = body  # original unformatted data
                self.subs = subs
                self.addl_rows = None
                self.format_section = format_section
                self.cur_page = cur_page
                self.prev_slice = {}
                self.this_slice = {}
                self.next_slice = {}

            def __call__(self, lines, cols, rows):
                self.append(lines, cols, rows)
                return self

            def __iter__(self, key: Union[slice, tuple, int] = None) -> Tuple:
                print("hit")
                if key:
                    if isinstance(key, tuple):
                        _slice = slice(*key)
                    else:
                        _slice = key
                else:
                    _slice = slice(0, len(self.lines), 1)

                for idx, (_lines, _rows, _cols) in enumerate(zip(self.lines[_slice],
                                                                 self.rows[_slice],
                                                                 self.cols[_slice])):
                    yield idx, _lines, _rows, _cols

            def append(self, lines: list = None, cols: int = None, rows: int = None):
                if lines:
                    self.lines += [lines]
                if cols:
                    self.cols += [cols]
                if rows:
                    self.rows += [rows]

            def get_cols(self, body_avail_rows: int = None):
                body_avail_rows = self.body_avail_rows if body_avail_rows is None else body_avail_rows
                # first pass to determine what is possible then break, second pass populates menu
                self.addl_rows = addl_rows  # FIXME why does addl_rows have value here
                for _pass in range(2):
                    col_lines = []
                    section_slices = {}
                    self.pages = {}
                    self.page = 1
                    _stop = False
                    item = 1
                    for idx, lines in enumerate(body):
                        rows = self.rows[idx]

                        if self.body_avail_rows - len(lines) >= 3 + addl_rows:
                            _end = self.body_avail_rows - len(lines) - addl_rows
                        else:
                            _end = self.body_avail_rows - addl_rows

                        # -- break sections with too many rows to fit tty into list of slices for each col that fit into tty
                        if 0 < _end < len(lines):
                            _slice = [slice(0, _end, 1)]
                            while len([lines[s] for s in _slice]) > self.body_avail_rows:
                                _start, _end = _end, _end + (self.body_avail_rows - addl_rows)
                                _slice += [slice(_start, _end, 1)]

                            # Add any remaining slices in section
                            if _end < rows:
                                _slice += [slice(_end, len(lines), 1)]
                        else:
                            _slice = [slice(0, len(lines), 1)]

                        for s in _slice:
                            sub_section = self.body[idx][s]

                            # -- Format Sub Heading for CONTINUED Lines
                            sub = None if not self.subs else self.subs[idx]
                            _sub_key = 0 if not reverse else -1
                            if sub and sub_section[_sub_key] != self.body[idx][0]:
                                sub = f"[CONTINUED] {sub.split('] ')[-1].split(' @')[0]}"

                            this_lines, this_cols, _ = self.format_section(
                                body=sub_section,
                                sub=sub,
                                index=item,
                                format_sub=format_subs,
                            )

                            item = item + len(sub_section) if not reverse else item - len(sub_section)

                            section_slices[idx] = s if not reverse else \
                                slice(len(self.body[idx]) - s.stop, len(self.body[idx]) - s.start)

                            if len(col_lines) + len(this_lines) > self.body_avail_rows:

                                if self.page not in self.pages:
                                    self.pager_write_first_col(col_lines, page=self.page)
                                else:
                                    self.pager_write_other_col(col_lines, page=self.page)

                                # if self.page == self.cur_page:
                                #     self.vert_cols += 1

                                if tty.cols and self.page_width + self.col_width > tty.cols:
                                    self.page += 1
                                    if _pass == 0:
                                        self.body_avail_rows -= 1
                                        _stop = True
                                        break

                                # -- Prev Col written update col with current lines
                                col_lines = this_lines
                                self.col_width = this_cols
                                if self.page == self.cur_page:
                                    self.this_slice = {**self.this_slice, **section_slices}
                                elif self.page < self.cur_page:
                                    self.prev_slice = {**self.prev_slice, **section_slices}
                                else:
                                    self.next_slice = {**self.next_slice, **section_slices}
                                section_slices = {}
                            else:  # -- Appending to Existing Column Col not written to Page Yet
                                if not reverse:
                                    col_lines += this_lines
                                else:
                                    col_lines = this_lines + col_lines

                                if this_cols > self.col_width:
                                    self.col_width = this_cols

                        # abort loop and start next loop with updated body size
                        if _stop:
                            break
                    # abort loop and start next loop with updated body size
                    if _stop:
                        break

                # Write Final Col
                if self.page not in self.pages:
                    self.pager_write_first_col(col_lines, page=self.page)  # type: ignore
                else:
                    self.pager_write_other_col(col_lines, page=self.page)  # type: ignore

                # if self.page == self.cur_page:
                #     self.vert_cols += 1

                if self.page == self.cur_page:
                    self.this_slice = {**self.this_slice, **section_slices}  # type: ignore
                elif self.page < self.cur_page:
                    self.prev_slice = {**self.prev_slice, **section_slices}  # type: ignore
                else:
                    self.next_slice = {**self.next_slice, **section_slices}  # type: ignore

            # -- // Helpers called by pager_write_col_to_page() \\ --
            def pager_write_first_col(self, col_lines: list, page: int) -> None:
                '''Writes first Col to Page

                Args:
                    col_lines (list): list of strings representing each menu option/choice.
                    page (int): The Page to write the output to
                '''
                self.pages[page] = [f"{line:{self.col_width}}" for line in col_lines]
                self.page_width = self.col_width
                self.prev_col_width = self.col_width

            def pager_write_other_col(self, col_lines: list, page: int) -> None:
                '''Writes additional col to existing page

                Args:
                    col_lines (list): list of strings representing each menu option/choice.
                    page (int): The Page to write the output to
                '''
                # - pad any cols that are shorter with spaces matching the longest col on the page
                if len(self.pages[page]) < len(col_lines):
                    for _ in range(len(col_lines) - len(self.pages[page])):
                        self.pages[page] += [" " * self.page_width]
                elif len(col_lines) < len(self.pages[page]):
                    for _ in range(len(self.pages[page]) - len(col_lines)):
                        col_lines += [f"{' ':{self.col_width}}"]

                col_pad = COL_PAD - self.left_offset
                if not reverse:
                    self.pages[page] = [
                        f"{line[0]}{' ':{col_pad}}{line[1]:{self.col_width}}"
                        for line in zip(self.pages[page], col_lines)
                    ]
                else:  # Add From Right to Left when parsing slices in reverse order (Back)
                    self.pages[page] = [
                        f"{line[0]:{self.col_width}}{' ':{col_pad}}{line[1]:{self.col_width}}"
                        for line in zip(col_lines, self.pages[page])
                    ]

                self.page_width += (col_pad + self.col_width)

        size = Size()
        start = item = 1
        addl_rows = 0
        for i, _section in enumerate(body):
            if by_tens and i > 0:
                item = start + 10 if item <= start + 10 else item
                start += 10
            this_lines, this_width, this_rows = self.format_section(
                body=_section,
                sub=None if subs is None else subs[i],
                index=item, format_sub=format_subs
            )
            if i == 0:
                addl_rows = this_rows - len(_section)
            size = size(this_lines, this_width, this_rows)
            item = item + len(_section)

        return size

    def toggle_legend(self):
        '''Toggles Display of the Legend.
        '''
        config.hide_legend = self.page.legend.hide = not self.page.legend.hide
        # if self.page.legend.hide:
        #     self.page.legend.lines = []
        #     self.page.legend.rows, self.page.legend.cols = 0, 0

    def pager_write_col_to_page(self, col_lines: list, col_slices: dict, legend_lines: list = None) -> None:
        '''Writes Column completed column to Page (Body) extending to legend once beyond legend text.

        Args:
            col_lines (list): list of formatted strings representing each menu option/choice
            col_slices (dict): {idx: (start, stop)} Where idx is the index of the parent section
                              from the original data provided to the menu (self.body_in) and
                              start, stop = where to slice that section for the current sub-section
            legend_lines (list, optional): list of formatted strings representing each menu option/choice
                                           written to legend once page is beyond legend text
                                           Default to None


        '''
        page = 1 if not self.pages else len(self.pages)
        if tty.cols and self.page_width + (COL_PAD - self.left_offset) + self.col_width > tty.cols:
            # Increment page unless this is first col of first page
            page = page if not self.pages[page] else page + 1

        for idx in col_slices:
            if page == self.cur_page:
                _attr = getattr(self.page, "this_slice")
            elif page < self.cur_page:
                _attr = getattr(self.page, "prev_slice")
            else:
                _attr = getattr(self.page, "next_slice")

            # pager only parses slices for next page adjust first slice to match index of original body (body_in)
            # if not self.cur_page < self.prev_page and idx in self.page.prev_slice:
            if not self.reverse:
                if idx in self.page.prev_slice:
                    col_slices[idx] = slice(
                        self.page.prev_slice[idx].stop + col_slices[idx].start,
                        self.page.prev_slice[idx].stop + col_slices[idx].stop,
                        1
                    )
            else:
                if idx in self.page.next_slice:
                    col_slices[idx] = slice(
                        self.page.next_slice[idx].stop - col_slices[idx].stop,
                        self.page.next_slice[idx].stop - col_slices[idx].start,
                        1
                    )

            if not _attr.get(idx):
                # overflow from prev_slices when parsing backward would go into next_slice, next_slice is retained
                # as prev this_slice so just ignore the overflow.
                # if not self.cur_page < self.prev_page or page <= self.cur_page:
                if not self.reverse or page <= self.cur_page:
                    _attr[idx] = col_slices[idx]
            else:
                # if not self.cur_page < self.prev_page:
                if not self.reverse:
                    _attr[idx] = slice(_attr[idx].start, col_slices[idx].stop, 1)
                else:
                    _attr[idx] = slice(col_slices[idx].start, _attr[idx].stop, 1)

        # -- // Col Full but PAGE DOESN't EXIST YET - Write First Col \\ --
        if not self.pages.get(page):
            # When parsing in reverse only print cur_page
            if not (page > self.cur_page and self.reverse):
                self.pager_write_first_col(col_lines, page)
        else:  # -- // PAGE NOT FULL write existing col_lines to page \\ --
            self.pager_write_other_col(col_lines, page)
            if legend_lines:
                self.pager_write_other_col(legend_lines, page)  # , legend=True)

        if page == self.cur_page:
            self.cur_page_width = self.page_width  # self.page_width updated in helper methods

    # -- // Helpers called by pager_write_col_to_page() \\ --
    def pager_write_first_col(self, col_lines: list, page: int) -> None:
        '''Writes first Col to Page

        Args:
            col_lines (list): list of strings representing each menu option/choice.
            page (int): The Page to write the output to
        '''
        self.pages[page] = [f"{line:{self.col_width}}" for line in col_lines]
        self.page_width = self.col_width
        self.prev_col_width = self.col_width

    def pager_write_other_col(self, col_lines: list, page: int) -> None:
        '''Writes additional col to existing page

        Args:
            col_lines (list): list of strings representing each menu option/choice.
            page (int): The Page to write the output to
        '''
        # - pad any cols that are shorter with spaces matching the longest col on the page
        if len(self.pages[page]) < len(col_lines):
            for _ in range(len(col_lines) - len(self.pages[page])):
                self.pages[page] += [" " * self.page_width]
        elif len(col_lines) < len(self.pages[page]):
            for _ in range(len(self.pages[page]) - len(col_lines)):
                col_lines += [f"{' ':{self.col_width}}"]

        col_pad = COL_PAD - self.left_offset
        if not self.reverse:
            self.pages[page] = [
                f"{line[0]}{' ':{col_pad}}{line[1]:{self.col_width}}"
                for line in zip(self.pages[page], col_lines)
            ]
        else:  # Add From Right to Left when parsing slices in reverse order (Back)
            self.pages[page] = [
                f"{line[0]:{self.col_width}}{' ':{col_pad}}{line[1]:{self.col_width}}"
                for line in zip(col_lines, self.pages[page])
            ]

        self.page_width += (col_pad + self.col_width)

    def pager_update_legend(self):
        '''Updates the Legend with Back/Next Options as appropriate for Paged menu.
        '''
        # -- Add Navigation options when menu is paged
        if self.page.prev_slice:
            self.actions = {**self.actions, **{"b": self.pager_prev_page}}
            if "back" not in self.page.legend.opts:
                self.page.legend.opts.insert(len(self.page.legend.opts) - 2, 'back')
            self.legend_options["back"] = ("b", "Back (Prev Page)")
        else:  # Page 1
            # Restore legend options to default (resets Back)
            self.legend_options = {**self.legend_options, **DEF_LEGEND_OPTIONS}
            # self.actions = {**self.actions, **self.menu_actions_in}
            if "b" in self.actions and self.actions["b"] and self.actions["b"].__name__ == "pager_prev_page":
                del self.actions["b"]
                if 'back' in self.page.legend.opts:
                    self.page.legend.opts.pop(
                        self.page.legend.opts.index('back')
                    )

        if self.page.next_slice:
            self.actions = {**self.actions, **{"n": self.pager_next_page}}
            self.page.legend.opts.insert(len(self.page.legend.opts) - 2, 'next')
        else:
            if "n" in self.actions:
                del self.actions["n"]
            if "next" in self.page.legend.opts:
                self.page.legend.opts.pop(
                    self.page.legend.opts.index('next')
                )

        # Tweak order of Final Legend Options so that last entries are refresh, back, next, exit
        for opt in ["refresh", "back", "next", "x"]:
            if opt in self.page.legend.opts:
                self.page.legend.opts.pop(
                    self.page.legend.opts.index(opt)
                )
                self.page.legend.opts += [opt]

    def pager_next_page(self):
        '''Configure Menu to advance to next Page.

        pbody and psubs attributes are updated, they override the original body/subs lists provided to the menu
        prev_slice is updated to include this_slice (the page your navigating from)
        this_slice is reset and populated (Slices for the page we are navigating to) by print_menu()
        next_slice is reset and populated with slices for anything that doesn't fit into this_slice (the new page)

        This ensures all sections print, if all pages where stashed on the original instantiation, a section could be
        skipped given the menu size is dynamic (log area, and user could resize)
        '''
        self.reverse = False
        self.pbody = [self.body_in[sec][self.page.next_slice[sec]] for sec in self.page.next_slice]
        self.pitems = [self.items_in[sec][self.page.next_slice[sec]] for sec in self.page.next_slice]
        self.psubs = None if not self.subs_in else [self.subs_in[sec] for sec in self.page.next_slice]

        # Add slices for this_page to prev_page adjust last section start/stop if it's split between pages
        _first_key = [_ for _ in sorted(self.page.this_slice.keys())][0]
        if _first_key in self.page.prev_slice:
            self.page.this_slice[_first_key] = slice(self.page.prev_slice[_first_key].start,
                                                     self.page.this_slice[_first_key].stop, 1)
        self.page.prev_slice = {**self.page.prev_slice, **self.page.this_slice}
        self.page.next_slice = {}

        self.prev_page = self.cur_page
        self.cur_page += 1

    def pager_prev_page(self, reprint: bool = False):
        '''Configure Menu to go to Prev Page.

        pbody and psubs attributes are updated, they override the original body/subs lists provided to the menu
        next_slice is set to this_slice (the page we are moving from)
        prev_slice remains as is (which will include all previous pages)
        this_slice is reset and populated by print_menu()

        The Menu is built in reverse, high to low / right to left

        This ensures all sections print, if all pages where stashed on the original instantiation, a section could be
        skipped given the menu size is dynamic (log area, and user could resize)
        '''
        if self.cur_page - 1 > 1 or reprint:
            self.reverse = True
            self.pbody = [self.body_in[sec][self.page.prev_slice[sec]][::-1]
                          for sec in sorted(self.page.prev_slice, reverse=True)]
            self.pitems = [self.items_in[sec][self.page.prev_slice[sec]][::-1]
                           for sec in sorted(self.page.prev_slice, reverse=True)]
            self.psubs = [self.subs_in[sec] for sec in sorted(self.page.prev_slice, reverse=True)]

            # Add slices for this_page to next page slices adjust last section start/stop if it's split between pages
            _last_key = [_ for _ in sorted(self.page.this_slice)][-1]
            if _last_key in self.page.next_slice:
                self.page.next_slice[_last_key] = slice(self.page.this_slice[_last_key].start,
                                                        self.page.next_slice[_last_key].stop, 1)
            self.page.next_slice = {**self.page.this_slice, **self.page.next_slice}

            self.prev_page = self.cur_page
        else:
            self.pbody, self.psubs, self.pitems = None, None, None
            self.prev_page = 1
            self.page.prev_slice = {}
            self.reverse = False
            self.actions = {**self.actions, **self.menu_actions_in}

        if not reprint:
            self.cur_page -= 1

    def pager_reverse_parse_update(self):
        # Clean prev_slices that have been populated to this_slice
        self.page.this_slice = {k: v for k, v in sorted(self.page.this_slice.items())}
        for idx in self.page.this_slice:
            if idx in self.page.prev_slice:
                if self.page.this_slice[idx].start == 0:  # _attr = this_slice
                    del self.page.prev_slice[idx]
                else:
                    self.page.prev_slice[idx] = slice(0, self.page.this_slice[idx].start, 1)
        # If user caused reduction of avail lines for error display or tty resize a section could be removed
        # this will re-populate any orphaned sections into prev_slice
        if len(self.page.prev_slice) not in self.page.this_slice:
            self.page.prev_slice[len(self.page.prev_slice)] = slice(0, len(self.body_in[len(self.page.prev_slice)]), 1)

    def update_slices(self):
        # Ensure slice dicts are sorted
        if self.page.prev_slice:
            self.page.prev_slice = {k: self.page.prev_slice[k] for k in sorted(self.page.prev_slice)}
        if self.page.this_slice:
            self.page.this_slice = {k: self.page.this_slice[k] for k in sorted(self.page.this_slice)}
        if self.page.next_slice:
            self.page.next_slice = {k: self.page.next_slice[k] for k in sorted(self.page.next_slice)}

        if not self.page.this_slice or self.cur_page > self.prev_page:
            return
        elif self.reverse:
            self.pager_reverse_parse_update()
        else:
            if self.page.next_slice:
                for _last_key in [_ for _ in self.page.this_slice.keys()][::-1]:
                    if _last_key not in self.page.next_slice:
                        break
                    else:
                        if self.page.this_slice[_last_key].stop == len(self.body_in[_last_key]):
                            del self.page.next_slice[_last_key]
                        else:
                            self.page.next_slice[_last_key] = slice(self.page.this_slice[_last_key].stop,
                                                                    self.page.next_slice[_last_key].stop, 1)
            if self.page.prev_slice:
                _first_key = [_ for _ in sorted(self.page.this_slice.keys())][-1]
                if _first_key in self.page.prev_slice:
                    if self.page.this_slice[_first_key].start == 0:
                        del self.page.prev_slice[_first_key]
                    else:
                        self.page.prev_slice[_first_key] = slice(0, self.page.this_slice[_first_key].start, 1)

# -- // MENU PART FORMATTERS \\ --
    @staticmethod
    def format_line(line: Union[str, bool]) -> object:
        return format_line(line)

    def format_header(self, text: Union[str, List[str]], width: int = MIN_WIDTH) -> MenuSection:
        if tty and width > tty.cols:
            width = tty.cols
        orig = text
        head_lines = ["=" * width]
        width_list = [width]
        text = utils.listify(text)
        for t in text:
            line = format_line(t)
            width_list += [line.len]
            a = width - line.len
            b = (a / 2) - 2
            c = int(b) if b == int(b) else int(b) + 1
            head_lines.append(f" {'-' * int(b)} {line.text} {'-' * c}")
        head_lines.append("=" * width)
        width_list += [width]

        _header = MenuSection(
            orig=orig,
            lines=head_lines,
            width_list=width_list,
            rows=len(head_lines),
            cols=max(width_list),
            name="header",
            update_method=self.format_header,
            update_args=(text, ),
            update_kwargs={"width": max(width_list)},
        )
        # _header.update_method = self.format_header
        # _header.update_args = (text)
        # _header.update_kwargs = {"width": width}
        self.page.header = _header

        return self.page.header

    def format_subhead(self, text: Union[str, list]) -> MenuSection:
        if not text:
            _subhead = MenuSection(name="subhead")
            self.page.subhead = _subhead
            return _subhead

        subhead = utils.listify(text)
        # indent each subhead line with a space if not provided that way
        subhead = [
            f"{' ' + line if not line.startswith(' ') else line}"
            for line in subhead
        ]
        subhead.insert(0, "")
        if not self.subs_in:
            subhead += [""]
        max_len = max([len(sh) for sh in subhead])

        _subhead = MenuSection(lines=subhead, rows=len(subhead), cols=max_len)
        _subhead.update_method = self.format_subhead
        _subhead.update_args = (text)
        self.page.subhead = _subhead

        return _subhead

    def format_section(self, body: list, sub: str = None, index: int = 1,
                       left_offset: int = None, format_sub: bool = False) -> MenuSection:
        left_offset = self.left_offset if not left_offset else left_offset
        max_len = 0
        mlines = []
        _end_index_len = len(str(len(body) + index - 1))
        indent = left_offset + _end_index_len + 2  # The 2 is for '. '
        width_list = []
        for _line in body:
            # -- format spacing of item entry --
            _i = f"{str(index)}. {' ' * (_end_index_len - len(str(index)))}"
            # -- generate line and calculate line length --
            _line = " " * left_offset + _i + _line.rstrip()  # rstrip handle errant tabs in dli port name
            line = format_line(_line)
            # if not self.cur_page < self.prev_page:
            if not self.reverse:
                width_list.append(line.len)
                mlines.append(line.text)
                index += 1
            else:
                width_list.insert(0, line.len)
                mlines.insert(0, line.text)
                index -= 1
        max_len = 0 if not width_list else max(width_list)
        if sub:
            # -- Add subs lines to top of menu item section --
            x = ((max_len - len(sub)) / 2) - (left_offset + (indent / 2))
            mlines.insert(0, "")
            width_list.insert(0, 0)
            if format_sub:
                mlines.insert(
                    1,
                    "{0}{1} {2} {3}".format(
                        " " * indent,
                        "-" * int(x),
                        sub,
                        "-" * int(x) if x == int(x) else "-" * (int(x) + 1),
                    ),
                )
                width_list.insert(1, len(mlines[1]))
            else:
                mlines.insert(1, " " * indent + sub)
                width_list.insert(1, len(mlines[1]))
                # update max_len in case subsheading is the longest line in the section
                max_len = max(width_list)
            mlines.insert(2, " " * indent + "-" * (max_len - indent))
            width_list.insert(2, len(mlines[2]))

        # -- adding padding to line to full width of longest line in section --
        mlines = [f"{line}{' ' * (max_len - line_len) if line_len < max_len else ''}"
                  for line, line_len in zip(mlines, width_list)]
        return mlines, max_len, len(mlines)

    def format_legend(self, legend: dict = {}, left_offset: int = None, **kwargs) -> MenuSection:
        if self.page.legend and self.page.legend.hide:
            return self.page.legend

        left_offset = self.left_offset if not left_offset else left_offset
        col_pad = COL_PAD - left_offset
        width_list = []
        opts = utils.listify(legend.get("opts", []))

        if kwargs.get("opts"):
            opts = [*opts, *utils.listify(opts)]

        # Ensure exit is the last option unless only 2 opts
        if "x" in opts:
            opts.pop(opts.index("x"))
        if len(opts) > 2:
            opts.append("x")
        else:
            opts.insert(0, "x")
        opts = utils.unique(opts)

        no_match_overrides, no_match_rjust = [], []  # init
        pre_text, post_text, legend_text = [], [], []  # init

        # replace any pre-defined options with those passed in as overrides
        legend_options = self.legend_options or DEF_LEGEND_OPTIONS
        if legend.get("overrides") and isinstance(legend["overrides"], dict):
            legend_options = {**legend_options, **legend["overrides"]}
            no_match_overrides = [
                e for e in legend["overrides"]
                if e not in legend_options and e not in legend.get("rjust", {})
            ]

        # update legend_options with any specially formmated (rjust) additions
        width = self.page.cols
        if legend.get("rjust"):
            r = legend.get("rjust")
            f = legend_options
            legend_overrides = {
                k: [f[k][0], "{}{}".format(f[k][1], r[k].rjust(width - len(f' {f[k][0]}.{" " if len(f[k][0]) == 2 else "  "}{f[k][1]}')))]
                for k in r
                if k in f
            }

            legend_options = {**legend_options, **legend_overrides}
            no_match_rjust = [e for e in legend["rjust"] if e not in legend_options]

        if legend.get("before"):
            legend["before"] = utils.listify(legend["before"])
            pre_text = [format_line(f" {line}") for line in legend["before"]]
            width_list += [line.len for line in pre_text]
            pre_text = [line.text for line in pre_text]

        if opts:
            f = legend_options
            legend_text = [
                f'{" " * left_offset}{f[k][0]}.{" " if len(f[k][0]) == 2 else "  "}{f[k][1]}'
                for k in opts
                if k in f
            ]
            legend_text = [format_line(line) for line in legend_text]
            legend_width_list = [line.len for line in legend_text]
            legend_text = [line.text for line in legend_text]
            _middle = int(len(legend_text) / 2)
            col_1_max = max([_len for _len in legend_width_list[_middle:len(legend_width_list)]])
            col_2_max = max([_len for _len in legend_width_list[0:_middle]] or [0])
            col_1_text = [line for line in legend_text[_middle:len(legend_text)]]
            col_2_text = [line for line in legend_text[0:_middle]]
            if col_2_max > 0:
                if len(col_2_text) < len(col_1_text):
                    col_2_text += [""]
                legend_text = [
                    f"{col[0]:{col_1_max}}{' ':{col_pad}}{col[1]:{col_2_max}}" for col in zip(col_1_text, col_2_text)
                ]
                width_list += [col_1_max + col_pad + col_2_max]
                legend_text.insert(0, " " + "-" * (col_1_max + col_pad + col_2_max - 1))

                # if multi-paged output, and there is room on last line of legend display toggle legend options
                if (self.actions and "b" in self.actions and self.actions["b"] and self.actions["b"].__name__ == "pager_prev_page") or 'next' in opts:
                    _tl = " 'TL' to hide "
                    line_slice = slice(0, len(legend_text[0]) - len(_tl) - 5)
                    legend_text[0] = f'{legend_text[0][line_slice]}{_tl}{"-" * 5}'

            else:  # legend consists of only 1 item
                legend_text = col_1_text
                width_list = legend_width_list
                legend_text.insert(0, " " + "-" * (col_1_max - 1))

        if legend.get("after"):
            legend["after"] = utils.listify(legend["after"])
            post_text = [format_line(f" {line}") for line in legend["after"]]
            width_list += [line.len for line in post_text]
            post_text = [line.text for line in post_text]

        mlines = [""] + pre_text + legend_text + post_text

        # log errors if non-match overrides/rjust options were sent
        if no_match_overrides + no_match_rjust:
            log.error(
                f'menu_formatting passed options ({",".join(no_match_overrides + no_match_rjust)})'
                " that lacked a match in legend_options = No impact to menu")

        # TODO update attributes if self.page has attr legend
        if not hasattr(self.page, "legend") or self.page.legend is None:
            _legend = MenuSection(lines=mlines, cols=max(width_list), rows=len(mlines), opts=opts or [],
                                  overrides=legend.get("overrides", {}), name="legend")
        else:
            _legend = getattr(self.page, "legend")
            _legend.lines = mlines
            _legend.cols = max(width_list)
            _legend.rows = len(mlines)
            _legend.opts = opts or []

        _legend.update_method = self.format_legend
        _legend.update_kwargs = {"legend": {"before": legend.get("before", []), "opts": _legend.opts,
                                            "after": legend.get("after", []), "overrides": _legend.overrides}}

        # _legend.update(width=self.page.cols)
        self.page.legend = _legend

        return _legend

    def format_footer(self, width: int = MIN_WIDTH) -> MenuSection:
        if tty and width > tty.cols:
            width = tty.cols
        foot_width_list = []
        mlines = []

        # Remove any error_msgs matching re from ignored_errors list
        if log.error_msgs:
            for _error in log.error_msgs:
                for e in self.ignored_errors:
                    _e = _error.strip("\r\n")
                    if hasattr(e, "match") and e.match(_e):
                        log.error_msgs.remove(_error)
                        break
                    elif isinstance(e, str) and e in _error:
                        log.error_msgs.remove(_error)
                        break

        # --// ERRORs - append to footer \\-- #
        if log.error_msgs:
            errors = log.error_msgs
            for _ in range(2):
                for _error in errors:
                    error = format_line(_error)
                    foot_width_list.append(error.len + 6)  # +6 is for "!! {error.msg} !!"
                    x = (width - (error.len + 6)) / 2
                    mlines.append(
                        "{0} {1}{2}{3} {0}".format(
                            self.log_sym_2bang,
                            " " * int(x),
                            error.text,
                            " " * int(x) if x == int(x) else " " * (int(x) + 1),
                        )
                    )

                if max(foot_width_list) <= width:
                    break
                else:
                    mlines = []
                    width = max(foot_width_list)

        mlines.insert(0, "")
        page_text = f"| Page {self.cur_page} of {len(self.pages)} |"

        top_line = "=" * width
        bot_line = top_line
        if self.page.prev_slice or self.page.next_slice:
            bot_line = "=" * (width - len(page_text) - 5) + page_text + "=" * 5
        if self.page.legend.hide:
            legend_text = " Use 'TL' to Toggle Legend "
            bot_line = "=" * 5 + legend_text + bot_line[len(legend_text) + 5:]

        if log.error_msgs:
            mlines.insert(1, top_line)

        mlines += [bot_line]

        _footer = MenuSection(
            lines=mlines,
            width_list=foot_width_list,
            rows=len(mlines),
            cols=width,
            update_method=self.format_footer,
            update_kwargs={"width": width},
            name="footer",
        )
        self.page.footer = _footer

        return _footer


# This object is currently not used, but will be to simplify the logic in cpi_exec
# Which currently is a functional but it's a Shit Show.
class MenuExecute:
    def __init__(self, function, args, kwargs, calling_menu: str = None):
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.calling_menu = calling_menu



#!/etc/ConsolePi/venv/bin/python3

from __future__ import annotations

import string
import subprocess
import shlex
from subprocess import TimeoutExpired
import time
import psutil
import os
import sys
import stat
import grp
import json
import threading
import socket
from io import StringIO
from halo import Halo
from typing import List, Dict

try:
    loc_user = os.getlogin()
except Exception:
    loc_user = os.getenv("SUDO_USER", os.getenv("USER"))


class Convert:
    def __init__(self, mac):
        self.orig = mac
        if not mac:
            mac = '0'
        self.clean = ''.join([c for c in list(mac) if c in string.hexdigits])
        self.ok = True if len(self.clean) == 12 else False
        self.cols = ':'.join(self.clean[i:i + 2] for i in range(0, 12, 2))
        self.dashes = '-'.join(self.clean[i:i + 2] for i in range(0, 12, 2))
        self.dots = '.'.join(self.clean[i:i + 4] for i in range(0, 12, 4))
        self.tag = f"ztp-{self.clean[-4:]}"
        self.dec = int(self.clean, 16) if self.ok else 0


class Mac(Convert):
    def __init__(self, mac):
        super().__init__(mac)
        oobm = hex(self.dec + 1).lstrip('0x')
        self.oobm = Convert(oobm)


class Utils:
    def __init__(self):
        self.Mac = Mac

    def user_input_bool(self, question):
        """Ask User Y/N Question require Y/N answer

        Error and reprompt if user's response is not valid
        Appends '? (y/n): ' to question/prompt provided

        Params:
            question:str, The Question to ask
        Returns:
            answer:bool, Users Response yes=True
        """
        valid_answer = ["yes", "y", "no", "n"]
        try:
            answer = input(question + "? (y/n): ").strip()
        except (KeyboardInterrupt, EOFError):
            print("")  # prevents header printing on same line when in debug
            return False
        while answer.lower() not in valid_answer:
            if answer != "":
                print(
                    f" \033[1;33m!!\033[0m Invalid Response '{answer}' Valid Responses: {valid_answer}"
                )
            answer = input(question + "? (y/n): ").strip()
        if answer[0].lower() == "y":
            return True
        else:
            return False

    def kill_hung_session(self, dev):
        """Kill hung picocom session

        If picocom process still active and user is diconnected from SSH
        it makes the device unavailable.  When error_handler determines that is the case
        This function is called giving the user the option to kill it.
        """

        def find_procs_by_name(name, dev):
            """Return the pid of process matching name, where dev was referenced in the cmdline options

            Params:
                name:str, name of process to find
                dev:str, dev which needs to be in the cmdline arguments for the process

            Returns:
                The pid of the matching process
            """
            ppid = None
            for p in psutil.process_iter(attrs=["name", "cmdline"]):
                if name == p.info["name"] and dev in p.info["cmdline"]:
                    ppid = p.pid  # if p.ppid() == 1 else p.ppid()
                    break
            return ppid

        def terminate_process(pid):
            """send terminate, then kill if still alive to pid

            params: pid of process to be killed
            """
            p = psutil.Process(pid)
            for x in range(0, 2):
                p.terminate()
                if p.status() != "Terminated":
                    p.kill()
                else:
                    break

        ppid = find_procs_by_name("picocom", dev)
        retry = 0
        msg = "\n{} appears to be in use (may be a previous hung session).\nDo you want to Terminate the existing session".format(
            dev.replace("/dev/", "")
        )
        if ppid is not None and self.user_input_bool(msg):
            while ppid is not None and retry < 3:
                print(
                    "An Existing session is already established to {}.  Terminating process {}".format(
                        dev.replace("/dev/", ""), ppid
                    )
                )
                try:
                    terminate_process(ppid)
                    time.sleep(3)
                    ppid = find_procs_by_name("picocom", dev)
                except PermissionError:
                    print(
                        "PermissionError: session is locked by user with higher priv. can not kill"
                    )
                    break
                except psutil.AccessDenied:
                    print(
                        "AccessDenied: Session is locked by user with higher priv. can not kill"
                    )
                    break
                except psutil.NoSuchProcess:
                    ppid = find_procs_by_name("picocom", dev)
                retry += 1
        return ppid is None

    def error_handler(self, cmd, stderr, user=loc_user):
        # TODO rather than running the command return something to the calling function instructing it to retry
        # with new cmd
        if isinstance(cmd, str):
            cmd = shlex.split(cmd)
        if stderr and "FATAL: cannot lock /dev/" not in stderr:
            # -- // KEY CHANGE ERROR \\ --
            if "WARNING: REMOTE HOST IDENTIFICATION HAS CHANGED!" in stderr:
                print(
                    "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"
                    "@    WARNING: REMOTE HOST IDENTIFICATION HAS CHANGED!     @\n"
                    "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"
                    "IT IS POSSIBLE THAT SOMEONE IS DOING SOMETHING NASTY!\n"
                    "Someone could be eavesdropping on you right now (man-in-the-middle attack)!\n"
                    "It is also possible that a host key has just been changed."
                )
                while True:
                    choice = ''
                    try:
                        choice = input(
                            "\nDo you want to remove the old host key and re-attempt the connection (y/n)? "
                        )
                        if choice.lower() in ["y", "yes"]:
                            _cmd = shlex.split(
                                stderr.replace("\r", "")
                                .split("remove with:\n")[1]
                                .split("\n")[0]
                                .replace("ERROR:   ", "")
                            )  # NoQA
                            _cmd = shlex.split("sudo -u {}".format(user)) + _cmd
                            subprocess.run(_cmd)
                            print("\n")
                            subprocess.run(cmd)
                            break
                        elif choice.lower() in ["n", "no"]:
                            break
                        else:
                            print(
                                "\n!!! Invalid selection {} please try again.\n".format(
                                    choice
                                )
                            )
                    except (KeyboardInterrupt, EOFError):
                        print("")
                        return "Aborted last command based on user input"
                    except ValueError:
                        print(
                            "\n!! Invalid selection {} please try again.\n".format(
                                choice
                            )
                        )
            elif (
                "All keys were skipped because they already exist on the remote system"
                in stderr
            ):
                return "Skipped - key already exists"

            elif "/usr/bin/ssh-copy-id: INFO:" in stderr:
                if "sh: 1:" in stderr:
                    return "".join(stderr.split("sh: 1:")[1:]).strip()
                else:
                    if '\n' in stderr:
                        stderr = stderr.replace('\r', '')
                        return "".join([line for line in stderr.split('\n')
                                        if not line.startswith('/usr/bin/ssh-copy-id: INFO:')
                                        ]).strip()

            # -- // SSH CIPHER SUITE ERRORS \\ --
            elif "no matching cipher found. Their offer:" in stderr:
                print("Connection Error: {}\n".format(stderr))
                cipher = stderr.split("offer:")[1].strip().split(",")
                aes_cipher = [c for c in cipher if "aes" in c]
                if aes_cipher:
                    cipher = aes_cipher[-1]
                else:
                    cipher = cipher[-1]
                cmd += ["-o", f"ciphers={cipher}"]

                print(f"Reattempting Connection using cipher {cipher}")
                # r = subprocess.run(cmd)
                # if r.returncode:
                #     time.sleep(3)
                #     return "Error on Retry Attempt"  # TODO better way... handle banners... paramiko?
                s = subprocess
                with s.Popen(cmd, stderr=s.PIPE, bufsize=1, universal_newlines=True) as p, StringIO() as buf1:
                    for line in p.stderr:
                        print(line, end="")
                        if not line.startswith('/usr/bin/ssh-copy-id: INFO:'):
                            buf1.write(line)

                    return buf1.getvalue()

            # TODO handle invalid Key Exchange -o KexAlgorithms=...
            else:
                return stderr  # return value that was passed in

        # Handle hung sessions always returncode=1 doesn't always present stderr
        elif cmd[0] == "picocom":
            if self.kill_hung_session(cmd[1]):
                subprocess.run(cmd)
            else:
                return "User Abort or Failure to kill existing session to {}".format(
                    cmd[1].replace("/dev/", "")
                )

    def shell_output_cleaner(self, output):
        strip_words = ["/usr/bin/ssh-copy-id: "]
        return "".join(
            [x.replace(i, "") for x in self.listify(output) for i in strip_words]
        )

    def do_shell_cmd(
        self,
        cmd,
        do_print=False,
        handle_errors=True,
        return_stdout=False,
        tee_stderr=False,
        timeout=5,
        shell=False,
        **kwargs,
    ):
        """Runs shell cmd (i.e. ssh) returns stderr if any by default

        Arguments:
            cmd {str|list} -- commands/args sent to subprocess

        Keyword Arguments:
            do_print {bool} -- Print stderr after cmd completes (default: {False})
            handle_errors {bool} -- Send stderr to error_handler (default: {True})
            return_stdout {bool} -- run with shell and return tuple returncode, stdout, stderr (default: {False})
            tee_stderr {bool} -- stderr is displayed and captured (default: {False})
            timeout {int} -- subprocess timeout (default: {5})
            shell {bool} -- run cmd as shell (default: {True})

        Returns:
            {str|tuple} -- By default there is no return unless there is an error.
            return_stdout=True will return tuple returncode, stdout, stderr
        """
        if return_stdout:
            res = subprocess.run(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
                **kwargs,
            )
            return res.returncode, res.stdout.strip(), res.stderr.strip()
        elif shell:
            res = subprocess.run(
                cmd,
                shell=True,
                stderr=subprocess.PIPE,
                universal_newlines=True,
                **kwargs,
            )
            if res.stderr:
                return (
                    res.stderr
                    if not handle_errors
                    else self.error_handler(cmd, res.stderr)
                )

        else:
            if isinstance(cmd, str):
                cmd = shlex.split(cmd)

            if tee_stderr:
                s = subprocess
                start_time = time.time()
                with s.Popen(
                    cmd, stderr=s.PIPE, bufsize=1, universal_newlines=True, **kwargs
                ) as p, StringIO() as buf1, StringIO() as buf2:
                    for line in p.stderr:
                        print(line, end="")
                        # handles login banners which are returned via stderr
                        if time.time() - start_time < timeout + 5:
                            buf1.write(line)
                        else:
                            buf2.write(line)

                    early_error = buf1.getvalue()
                    late_error = buf2.getvalue()

                # if connections lasted 20 secs past timeout assume the early stuff was innocuous
                if time.time() - start_time > timeout + 10:
                    error = late_error
                else:
                    error = early_error

                if handle_errors and error and p.returncode != 0:
                    error = self.error_handler(cmd, error)
                else:
                    error = None
                return error
            else:
                proc = subprocess.Popen(
                    cmd, stderr=subprocess.PIPE, universal_newlines=True, **kwargs
                )

                try:
                    err = proc.communicate(timeout=timeout)[1]
                    if err is not None and do_print:
                        print(self.shell_output_cleaner(err), file=sys.stdout)
                    # if proc.returncode != 0 and handle_errors:
                    if err and handle_errors:
                        err = self.error_handler(cmd, err)

                    proc.wait()
                    return err
                except (TimeoutExpired, TimeoutError):
                    return "Timed Out.  Host is likely unreachable."

    def check_install_apt_pkg(self, pkg: str, verify_cmd=None):
        verify_cmd = "which {}".format(pkg) if verify_cmd is None else verify_cmd

        def verify(verify_cmd):
            if isinstance(verify_cmd, str):
                verify_cmd = shlex.split(verify_cmd)
            r = subprocess.run(
                verify_cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE
            )
            return r.returncode == 0

        if not verify(verify_cmd):
            resp = subprocess.run(
                ["/bin/bash", "-c", "apt install -y {}".format(pkg)],
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )
            if resp != 0:
                if verify(verify_cmd):
                    return (0, "{} Install Success".format(pkg))
        else:
            return (0, "Already Installed")

        return (
            resp.returncode,
            resp.stdout.decode("UTF-8")
            if resp.returncode == 0
            else resp.stderr.decode("UTF-8"),
        )

    def set_perm(self, file, user: str = None, group: str = None, other: str = None):
        _modes = {
            'user': {    # -- user and group set by default for our purposes --
                'r': 0,  # stat.S_IRUSR,
                'w': 0,  # stat.S_IWUSR,
                'x': stat.S_IXUSR
            },
            'group': {
                'r': 0,  # stat.S_IRGRP,
                'w': 0,  # stat.S_IWGRP,
                'x': stat.S_IXGRP
            },
            'other': {
                'r': stat.S_IROTH,
                'w': stat.S_IWOTH,
                'x': stat.S_IXOTH
            }
        }
        # -- by default set group ownership to consolepi and rw for user and group
        gid = grp.getgrnam("consolepi").gr_gid
        if os.geteuid() == 0:
            os.chown(file, 0, gid)
            _perms = stat.S_IWGRP + stat.S_IRGRP + stat.S_IWUSR + stat.S_IRUSR
            for k, v in [('user', user), ('group', group), ('other', other)]:
                if v:
                    # remove _perms added by this helper func by default if provided
                    if k and (k == 'user' or k == 'group'):
                        v = v.replace('r', '').replace('w', '')
                    for _m in list(v.lower()):
                        if not (_m == 'x' and os.path.isdir(file)):
                            # print(f"Add {k}: {_m}")
                            _perms += _modes[k][_m]

            # set x for all when dir (allow cd to the dir)
            if os.path.isdir(file):
                # print("add x for all")
                _perms += stat.S_IXUSR + stat.S_IXGRP + stat.S_IXOTH

            # if other and 'r' in other:
            #     _perms += stat.S_IROTH
            os.chmod(file, (_perms))

    def json_print(self, obj):
        print(json.dumps(obj, indent=4, sort_keys=True))

    def format_eof(self, file):
        cmd = "sed -i -e :a -e {} {}".format("'/^\\n*$/{$d;N;};/\\n$/ba'", file)
        return self.do_shell_cmd(cmd, handle_errors=False)

    def append_to_file(self, file, line):
        """Determine if last line of file includes a newline character
        write line provided on the next line

        params:
            file: file to write to
            line: line to write

        Returns:
            str if an error occurs otherwise Nothing
        """
        # strip any blank lines from end of file and remove LF
        self.format_eof(file)

        try:
            # determine if last line in file has LF
            with open(file) as f:
                _lines = f.readlines()
                _last = _lines[-1]

            # Make sure blank line will be placed at EOF and
            # prepend newline if last line of file lacks it to prevent appending to that line
            line = f"{line}\n" if not line.endswith("\n") else line
            line = f"\n{line}" if not _last.endswith("\n") else line

            with open(file, "a+") as f:
                f.write(line)
        except Exception as e:
            return e

    def get_picocom_ver(self):
        """return version of picocom"""
        x = subprocess.run(
            "which picocom >/dev/null && picocom  --help | head -1 | cut -dv -f2",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
        )
        x = x.stdout.decode("UTF-8").strip()
        if not x:
            # print("\nConsolePi Menu Requires picocom which doesn't appear to be installed")
            # print("Install with 'sudo apt install picocom'")
            # sys.exit(1)
            return 0
        else:
            return float(x)

    def get_ser2net_ver(self) -> str | None:
        """return version of ser2net"""
        x = subprocess.run(
            "which ser2net >/dev/null && ser2net -v | cut -d' ' -f3",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
        )
        x = x.stdout.decode("UTF-8").strip()

        return "" if not x else x

    def verify_telnet_installed(self, host_dict):
        """Install TELNET pkg if not already and TELNET hosts are defined

        params: host_dict.  The dict with manually defined host details
        return: True if not needed by any defined hosts, or is installed.
            returns are not currently used.
        """

        def telnet_install_thread(host_dict=host_dict):
            if "telnet" not in host_dict["_methods"]:
                return True

            r = self.check_install_apt_pkg("telnet")
            if r[0] != 0:
                # apt seems to return an error even when success
                r = subprocess.run(["which", "telnet"], capture_output=True).returncode
                if r != 0:
                    return False
            else:
                return True

        threading.Thread(
            target=telnet_install_thread, name="telnet_install_verify"
        ).start()

    def get_tty_size(self):
        size = subprocess.run(["stty", "size"], stdout=subprocess.PIPE)
        rows, cols = size.stdout.decode("UTF-8").split()
        return int(rows), int(cols)

    # TODO check if using kwarg sort added to fix re-order of error_msgs
    def unique(self, _list, sort=False):
        out = []
        [out.append(i) for i in _list if i not in out and i is not None]
        return out if not sort else sorted(out)

    def is_reachable(self, host, port, timeout=3, silent=False):
        s = socket.socket()
        try:
            s.settimeout(timeout)
            s.connect((host, port))
            _reachable = True
        except Exception as e:
            if not silent:
                print("something's wrong with %s:%d. Exception is %s" % (host, port, e))
            _reachable = False
        finally:
            s.close()
        return _reachable

    def format_dev(self, dev, hosts=None, outlet_groups: List[str] = None, udev=None, with_path=False) -> Dict[str, int | List[int]]:
        """Properly format devs found in user created JSON

        Args:
            dev (str | List[str]): An individual device or list of devices to be formatted (i.e. '/dev/idf_switch_1)
            hosts (dict, optional): Manually defined hosts from config. Defaults to None.
            outlet_groups (List[str], optional): Outlet Groups defined in the config. Defaults to None.
            udev (_type_, optional): _description_. Defaults to None.
            with_path (bool, optional): Strips any prefix if False adds /dev/, /host/, /group/ if True. Defaults to False.

        Returns:
            Dict[str, int | List[int]]: Linked Devices dictionary
        """
        if udev and not hosts:  # TODO don't think udev is ever passed to this func
            data = udev
            pfx = "/dev/"
            pfx_else = "/host/"
            print("DEBUG HIT HERE REMOVE THIS")
        else:
            data = hosts.get("_host_list", [])
            pfx = "/host/"
            pfx_else = "/dev/"

        dev = [dev] if not isinstance(dev, (list, dict)) else dev
        # dev_out = {}
        if with_path:
            if isinstance(dev, list):
                # d1 = [pfx + d for d in dev if '/' not in d and pfx + d in data]
                d_out = [
                    f"{f'{pfx}{d}' if f'{pfx}{d}' in data else None or f'/group/{d}' if d in outlet_groups else None or f'{pfx_else}{d}'}"
                    for d in dev
                    if "/" not in d
                ]
            elif isinstance(dev, dict):
                ...
                d_out = {
                    f"{f'{pfx}{d}' if f'{pfx}{d}' in data else None or f'/group/{d}' if d in outlet_groups else None or f'{pfx_else}{d}'}": dev[d]
                    for d in dev
                    if "/" not in d
                }
            else:
                d_out = None
            # for d in dev:
            #     if '/' not in d:
            #         _pfx = pfx if pfx + d in data else pfx_else
            #         if isinstance(dev, list):  # dli represented by dict dev: port or port list
            #             dev[dev.index(d)] = _pfx + d
            #         else:
            #             dev_out[_pfx + d] = dev[d]
            return d_out
        else:  # return dev list with prefixes stripped
            return (
                [d.split("/")[-1] for d in dev]
                if isinstance(dev, list)
                else {d.split("/")[-1]: dev[d] for d in dev}
            )

    def valid_file(self, filepath):
        return os.path.isfile(filepath) and os.stat(filepath).st_size > 0

    def listify(self, var):
        if var is None:
            return []
        else:
            return var if isinstance(var, list) else [var]

    def get_host_short(self, host):
        """Extract hostname from fqdn

        Arguments:
            host {str} -- hostname. If ip address is provided it's returned as is

        Returns:
            str -- host_short (lab1.example.com becomes lab1)
        """
        return (
            host.split(".")[0]
            if "." in host and not host.split(".")[0].isdigit()
            else host
        )

    def spinner(self, spin_txt, function, *args, **kwargs):
        spinner = kwargs.get("spinner", "dots")
        if sys.stdin.isatty():
            with Halo(text=spin_txt, spinner=spinner):
                return function(*args, **kwargs)



#!/etc/ConsolePi/venv/bin/python3

import sys
import subprocess

sys.path.insert(0, '/etc/ConsolePi/src/pypkg')
from consolepi import config, utils, log  # type: ignore # NoQA
from consolepi.consolepi import ConsolePi  # type: ignore # NoQA


cpi = ConsolePi()
local = cpi.local
hostname = local.hostname
adapters = local.adapters
interfaces = local.interfaces
remotes = cpi.remotes.data
details = local.data
dump = {'local': details, 'remotes': config.remotes}


def jprint(data):
    utils.json_print(data)
    if isinstance(data, dict):
        if len(data) > 1:
            _keys = [k.replace("/dev/", "") for k in data.keys()]
            _len = len(_keys)
            print(f"\n--\n{_keys}\nTotal: {_len}\n--")


def get_outlets():
    if not config.outlets:
        cpi.cpiexec.pwr_init_complete = True
    if config.outlets and not cpi.cpiexec.pwr_init_complete:
        utils.spinner('Waiting for Power Threads To Complete', cpi.cpiexec.wait_for_threads)
        if cpi.pwr:
            return cpi.pwr.data


if config.remotes:
    for r in config.remotes:
        if r not in cpi.remotes.data:
            dump['remotes'][r]['!! WARNING !!'] = 'This Device is Currently Unreachable'

if len(sys.argv) > 1:
    if sys.argv[1] == 'adapters':
        jprint(adapters)
    elif sys.argv[1] == 'interfaces':
        jprint(interfaces)
    elif sys.argv[1] == 'outlets':
        jprint(get_outlets())
    elif sys.argv[1] == 'hosts':
        _ = subprocess.run("sed -n '/HOSTS:/,/^ *$/p' /etc/ConsolePi/ConsolePi.yaml | more", shell=True)
        print(f"Total: {len(config.hosts.get('_host_list', []))}")
    elif sys.argv[1] == 'remotes':
        if len(sys.argv) == 2:
            jprint(remotes)
        elif len(sys.argv) == 4:
            if sys.argv[2] == 'del':
                if sys.argv[3] in remotes:
                    print('Removing ' + sys.argv[3] + ' from local cloud cache')
                    remotes.pop(sys.argv[3])
                    cpi.remotes.update_local_cloud_file(remote_consoles=remotes, current_remotes=remotes)
                    print('Remotes remaining in local cache')
                    jprint(remotes)
                    print('{} Removed from local cache'.format(sys.argv[3]))
                else:
                    print(f'!! {sys.argv[3]} Not Found in Local Cache')
    elif sys.argv[1] == 'local':
        jprint(details)
    else:
        try:
            print(f'\nRemote Cache data for {sys.argv[1]}:')
            jprint(remotes[sys.argv[1]])
        except KeyError:
            dump['outlets'] = get_outlets()
            jprint(dump)
            print('!!!  {} Not Found entire data set above   !!!'.format(sys.argv[1]))

else:
    dump['outlets'] = get_outlets()
    jprint(dump)

if log.error_msgs:
    print('!! ', end='')
    print('\n!! '.join(log.error_msgs))

if config.outlets and not cpi.cpiexec.pwr_init_complete:
    utils.spinner('Exiting... Waiting for Power Threads To Complete', cpi.cpiexec.wait_for_threads)



#!venv/bin/python3

from waitress import serve
from flask import Flask, jsonify, request
from consolepi.common import ConsolePi_data
from time import time

app = Flask(__name__)
config = ConsolePi_data(do_print=False)
# outlets = config.outlets
log = config.log
user = config.USER # pylint: disable=maybe-no-member
last_update = int(time())

def log_request(route):
    log.info('[API RQST IN] {} Requesting -- {} -- Data via API'.format(request.remote_addr, route))

@app.route('/api/v1.0/adapters', methods=['GET'])
def get_adapters():
    log_request('adapters')
    global last_update
    # if data has been refreshed in the last 20 seconds trust it is valid
    # prevents multiple simul calls to get_local after mdns_refresh and
    # subsequent API calls from all other ConsolePi on the network
    if int(time()) - last_update > 20:
        last_update = int(time())
        return jsonify({'adapters': config.get_adapters(do_print=False)})
    else:
        return jsonify({'adapters': config.adapters})

@app.route('/api/v1.0/remotes', methods=['GET'])
def get_cache():
    log_request('remotes')
    return jsonify({'remotes': config.get_local_cloud_file()})

@app.route('/api/v1.0/interfaces', methods=['GET'])
def get_ifaces():
    log_request('ifaces')
    return jsonify({'interfaces': config.get_if_ips()})

@app.route('/api/v1.0/outlets', methods=['GET'])
def get_outlets():
    log_request('outlets')
    # -- Collect Outlet Details remove sensitive data --
    outlets = config.pwr.pwr_get_outlets()
    if outlets and 'linked' in outlets:
        for grp in outlets['linked']:
            for x in ['username', 'password']:
                if x in outlets['linked'][grp]:
                    del outlets['linked'][grp][x]
    return jsonify(outlets)

@app.route('/api/v1.0/details', methods=['GET'])
def get_details():
    log_request('details')
    return jsonify({config.hostname: {'adapters': config.get_adapters(), 'interfaces': config.get_if_ips(), 'user': user}})

if __name__ == '__main__':
    serve(app, host='0.0.0.0', port=5000)



#!/etc/ConsolePi/venv/bin/python3

from zeroconf import ServiceInfo, Zeroconf
import time
import json
import socket
import pyudev
import threading
import struct
import sys
import setproctitle

from rich.traceback import install
install(show_locals=True)

sys.path.insert(0, '/etc/ConsolePi/src/pypkg')
from consolepi import log, config  # type: ignore # NoQA
from consolepi.consolepi import ConsolePi  # type: ignore # NoQA
from consolepi.gdrive import GoogleDrive  # type: ignore # NoQA


UPDATE_DELAY = 30

setproctitle.setproctitle("consolepi-mdnsreg")


class MDNS_Register:

    def __init__(self):
        self.zeroconf = Zeroconf()
        self.context = pyudev.Context()
        self.cpi = ConsolePi()

    def build_info(self, squash=None, local_adapters=None):
        local = self.cpi.local
        local.data = local.build_local_dict(refresh=True)
        loc = local.data[local.hostname]
        loc['hostname'] = local.hostname
        for a in loc['adapters']:
            if 'udev' in loc['adapters'][a]:
                del loc['adapters'][a]['udev']

        # ip_w_gw = loc['interfaces'].get('_ip_w_gw', '127.0.0.1')

        # if data set is too large for mdns browser on other side will retrieve via API
        if squash is not None:
            if squash == 'interfaces':
                del loc['adapters']
                x = loc['interfaces']
                loc['interfaces'] = {k: {i: v[i] for i in v if i not in ['mac', 'isgw']}
                                     for k, v in x.items() if '.' not in k and not k.startswith('_')}
            else:
                del loc['adapters']

        log.debug('[MDNS REG] Current content of local_data \n{}'.format(json.dumps(loc, indent=4, sort_keys=True)))
        loc = {k: '{}'.format(loc[k] if isinstance(loc[k], str) else json.dumps(loc[k])) for k in loc.keys()}

        info = ServiceInfo(
            "_consolepi._tcp.local.",
            local.hostname + "._consolepi._tcp.local.",
            addresses=[socket.inet_aton(ip) for ip in local.get_ip_list()],
            port=config.api_port,
            properties=loc,
            server=f'{local.hostname}.local.'
        )

        return info

    def update_mdns(self, device=None, action=None, *args, **kwargs):
        zeroconf = self.zeroconf
        info = self.try_build_info()

        def sub_restart_zc():
            log.info('[MDNS REG] mdns_refresh thread Start... Delaying {} Seconds'.format(UPDATE_DELAY))
            time.sleep(UPDATE_DELAY)  # Wait x seconds and then update, to accomodate multiple add removes
            zeroconf.update_service(info)
            zeroconf.unregister_service(info)
            time.sleep(5)
            zeroconf.register_service(info)
            log.info('[MDNS REG] mdns_refresh thread Completed')

        if device is not None:
            abort_mdns = False
            for thread in threading.enumerate():
                if 'mdns_refresh' in thread.name:
                    log.debug('[MDNS REG] Another mdns_refresh thread already queued, this thread will abort')
                    abort_mdns = True
                    break

            if not abort_mdns:
                threading.Thread(target=sub_restart_zc, name='mdns_refresh', args=()).start()
                log.debug('[MDNS REG] mdns_refresh Thread Started.  Current Threads:\n    {}'.format(threading.enumerate()))

            log.info('[MDNS REG] detected change: {} {}'.format(device.action, device.sys_name))
            if config.cloud:     # pylint: disable=maybe-no-member
                abort = False
                for thread in threading.enumerate():
                    if 'cloud_update' in thread.name:
                        log.debug('[MDNS REG] Another cloud Update thread already queued, this thread will abort')
                        abort = True
                        break

                if not abort:
                    threading.Thread(target=self.trigger_cloud_update, name='cloud_update', args=()).start()
                    log.debug('[MDNS REG] Cloud Update Thread Started.  Current Threads:\n    {}'.format(threading.enumerate()))

    def try_build_info(self):
        # Try sending with all data
        local = self.cpi.local

        info = ServiceInfo(
            "_consolepi._tcp.local.",
            local.hostname + "._consolepi._tcp.local.",
            server=f'{local.hostname}.local.'
        )

        try:
            info = self.build_info()
        except (struct.error, ValueError) as e:
            log.debug('[MDNS REG] data is too big for mdns, removing adapter data \n    {} {}'.format(e.__class__.__name__, e))
            log.debug('[MDNS REG] offending payload \n    {}'.format(json.dumps(local.data, indent=4, sort_keys=True)))
            # Too Big - Try sending without adapter data
            try:
                info = self.build_info(squash='adapters')
            except (struct.error, ValueError) as e:
                log.warning('[MDNS REG] data is still too big for mdns, reducing interface payload \n'
                            '    {} {}'.format(e.__class__.__name__, e))
                log.debug('[MDNS REG] offending interface data \n    {}'.format(
                          json.dumps(local.interfaces, indent=4, sort_keys=True)))
                try:
                    info = self.build_info(squash='interfaces')
                except (struct.error, ValueError):
                    log.critical('[MDNS REG] data is still too big for mdns')
                    log.debug(
                        '[MDNS REG] offending interface data \n    {}'.format(
                            json.dumps(local.interfaces, indent=4, sort_keys=True)
                        )
                    )

        return info

    def trigger_cloud_update(self):
        local = self.cpi.local
        remotes = self.cpi.remotes
        log.info('[MDNS REG] Cloud Update triggered delaying {} seconds'.format(UPDATE_DELAY))
        time.sleep(UPDATE_DELAY)  # Wait 30 seconds and then update, to accomodate multiple add removes
        data = local.build_local_dict(refresh=True)
        for a in local.data[local.hostname].get('adapters', {}):
            if 'udev' in local.data[local.hostname]['adapters'][a]:
                del local.data[local.hostname]['adapters'][a]['udev']

        log.debug(f'[MDNS REG] Final Data set collected for {local.hostname}: \n{json.dumps(data)}')

        remote_consoles = {}
        if config.cloud_svc == 'gdrive':  # pylint: disable=maybe-no-member
            cloud = GoogleDrive(local.hostname)

            remote_consoles = cloud.update_files(data)

        # Send remotes learned from cloud file to local cache
        if len(remote_consoles) > 0 and 'Gdrive-Error' not in remote_consoles:
            remotes.update_local_cloud_file(remote_consoles)
            log.info('[MDNS REG] Cloud Update Completed, Found {} Remote ConsolePis'.format(len(remote_consoles)))
        else:
            log.warning('[MDNS REG] Cloud Update Completed, No remotes found, or Error Occured')

    def run(self):
        zeroconf = self.zeroconf
        info = self.try_build_info()

        zeroconf.register_service(info)
        # monitor udev for add/remove of usb-serial adapters
        monitor = pyudev.Monitor.from_netlink(self.context)
        monitor.filter_by('usb')
        observer = pyudev.MonitorObserver(monitor, name='udev_monitor', callback=self.update_mdns)
        observer.start()
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            pass
        finally:
            print("Unregistering...")
            zeroconf.unregister_service(self.build_info())
            zeroconf.close()
            observer.send_stop()


if __name__ == '__main__':
    mdnsreg = MDNS_Register()
    mdnsreg.run()



#!/etc/ConsolePi/venv/bin/python3
'''
Convert ConsolePi system file template to final state

params:
argv[1] template name without j2 extension (file should be in /etc/ConsolePi/src/j2)
argv[2] ignored by this script used by installer which compares resulting template to file @ argv[2] and updates if different
remaining arguments (3+) should be in form template_var=desired_value

Output:
Creates file in /tmp/ directory (/tmp/argv[1]) used to compare against system file specified in argv[2] by install script.
'''
import jinja2
import sys


def parse_args():
    parameter_dict = {}
    for user_input in sys.argv[3:]:
        if "=" not in user_input:
            continue
        varname = user_input.split("=")[0]
        varvalue = '='.join(user_input.split("=")[1:])
        varvalue = varvalue.replace('{{cr}}', '\n')
        if varvalue.lower() in ["true", "false"]:
            varvalue = True if varvalue.lower() == "true" else False
        parameter_dict[varname] = varvalue
    return parameter_dict


parameter_dict = parse_args()
templateLoader = jinja2.FileSystemLoader(searchpath="/etc/ConsolePi/src/j2/")
templateEnv = jinja2.Environment(loader=templateLoader, keep_trailing_newline=True, trim_blocks=True, lstrip_blocks=True)
template = templateEnv.get_template(sys.argv[1] + '.j2')
outputText = template.render(parameter_dict)
template.stream(parameter_dict).dump(f'/tmp/{sys.argv[1]}')

# print(outputText)



#!/etc/ConsolePi/venv/bin/python3

""" Browse for other ConsolePis on the network
"""

import json
import time
import sys
from zeroconf import ServiceBrowser, ServiceStateChange, Zeroconf
import setproctitle
import asyncio

from rich.traceback import install
install(show_locals=True)

sys.path.insert(0, '/etc/ConsolePi/src/pypkg')
from consolepi import log, config  # type: ignore # NoQA
from consolepi.consolepi import ConsolePi  # type: ignore # NoQA


RESTART_INTERVAL = 300  # time in seconds browser service will restart

setproctitle.setproctitle("consolepi-mdnsbrowser")


class MDNS_Browser:

    def __init__(self, show=False):
        config.cloud = False  # mdns doesn't need to sync w cloud
        self.cpi = ConsolePi(bypass_outlets=True, bypass_cloud=True)
        self.debug = config.cfg.get('debug', False)
        self.show = show
        self.stop = False
        self.discovered = []    # for display when running interactively, resets @ every restart
        self.d_discovered = []  # used when running as daemon (doesn't reset)
        self.no_adapters = []  # If both mdns and API report no adapters for remote add to list to prevent subsequent API calls
        self.startup_logged = False
        self.zc = Zeroconf()

    def on_service_state_change(self,
                                zeroconf: Zeroconf, service_type: str, name: str, state_change: ServiceStateChange) -> None:
        if self.cpi.local.hostname == name.split(".")[0]:
            return
        if state_change is not ServiceStateChange.Added:
            return

        info = zeroconf.get_service_info(service_type, name)
        if not info:
            log.warning(f'[MDNS DSCVRY] {name}: No info found')
            return
        if not hasattr(info, "properties") or not info.properties:
            log.warning(f'[MDNS DSCVRY] {name}: No properties found')
            return

        properties = info.properties

        cpi = self.cpi
        mdns_data = None
        update_cache = False
        try:
            mdns_data = {
                k.decode('UTF-8'):
                v.decode('UTF-8') if len(v) == 0 or not v.decode('UTF-8')[0] in ['[', '{'] else json.loads(v.decode('UTF-8'))  # NoQA
                for k, v in properties.items()
            }
        except Exception as e:
            log.exception(
                f"[MDNS DSCVRY] {e.__class__.__name__} occured while parsing mdns_data:\n {mdns_data}\n"
                f"Exception: \n{e}"
            )
            log.error(f"[MDNS DSCVRY] entry from {name} ignored due to parsing exception.")
            return

        hostname = mdns_data.get('hostname')
        interfaces = mdns_data.get('interfaces', [])

        log_out = json.dumps(mdns_data, indent=4, sort_keys=True)
        log.debug(f'[MDNS DSCVRY] {hostname} Properties Discovered via mdns:\n{log_out}')

        rem_ip = mdns_data.get('rem_ip')
        if not rem_ip:
            if len(mdns_data.get('interfaces', [])) == 1:
                rem_ip = [interfaces[i]['ip'] for i in interfaces]
                rem_ip = rem_ip[0]
            else:
                rem_ip = None if hostname not in cpi.remotes.data or 'rem_ip' not in cpi.remotes.data[hostname] \
                    else cpi.remotes.data[hostname]['rem_ip']

        cur_known_adapters = cpi.remotes.data.get(hostname, {'adapters': None}).get('adapters')

        # -- Log new entry only if this is the first time it's been discovered --
        if hostname not in self.d_discovered:
            self.d_discovered += [hostname]
            log.info('[MDNS DSCVRY] {}({}) Discovered via mdns'.format(
                hostname, rem_ip if rem_ip is not None else '?'))

        from_mdns_adapters = mdns_data.get('adapters')
        mdns_data['rem_ip'] = rem_ip
        mdns_data['adapters'] = from_mdns_adapters if from_mdns_adapters else cur_known_adapters
        mdns_data['source'] = 'mdns'
        mdns_data['upd_time'] = int(time.time())
        mdns_data = {hostname: mdns_data}

        # update from API only if no adapter data exists either in cache or from mdns that triggered this
        # adapter data is updated on menu_launch either way
        if (not mdns_data[hostname]['adapters'] and hostname not in self.no_adapters) or \
                hostname not in cpi.remotes.data:
            log.info(f"[MDNS DSCVRY] {info.server.split('.')[0]} provided no adapter data Collecting via API")
            # TODO check this don't think needed had a hung process on one of my Pis added it to be safe
            try:
                # TODO we are setting update time here so always result in a cache update with the restart timer
                res = asyncio.run(cpi.remotes.api_reachable(hostname, mdns_data[hostname]))
                update_cache = res.update
                if not res.data.get('adapters'):
                    self.no_adapters.append(hostname)
                elif hostname in self.no_adapters:
                    self.no_adapters.remove(hostname)
                mdns_data[hostname] = res.data
            except Exception as e:
                log.exception(f'Exception occurred verifying reachability via API for {hostname}:\n{e}')

        if self.show:
            if hostname in self.discovered:
                self.discovered.remove(hostname)
            self.discovered.append('{}{}'.format(hostname, '*' if update_cache else ''))
            print(hostname + '({}) Discovered via mdns.'.format(rem_ip if rem_ip is not None else '?'))

            try:
                print(
                    '{}\n{}\n{}'.format(
                        'mdns: None' if from_mdns_adapters is None else 'mdns: {}'.format(
                            [d.replace('/dev/', '') for d in from_mdns_adapters]
                            if not isinstance(from_mdns_adapters, list) else
                            [d['dev'].replace('/dev/', '') for d in from_mdns_adapters]
                        ),
                        'api (mdns trigger): None' if not mdns_data[hostname]['adapters'] else 'api (mdns trigger): {}'.format(
                            [d.replace('/dev/', '') for d in mdns_data[hostname]['adapters']]
                            if not isinstance(mdns_data[hostname]['adapters'], list) else
                            [d['dev'].replace('/dev/', '') for d in mdns_data[hostname]['adapters']]
                        ),
                        'cache: None' if cur_known_adapters is None else 'cache: {}'.format(
                            [d.replace('/dev/', '') for d in cur_known_adapters]
                            if not isinstance(cur_known_adapters, list) else
                            [d['dev'].replace('/dev/', '') for d in cur_known_adapters]
                        )
                    )
                )
            except TypeError as e:
                print(f'EXCEPTION: {e}')

            print(f'\nDiscovered ConsolePis: {self.discovered}')
            print("press Ctrl-C to exit...\n")

        log.debugv(
            f"[MDNS DSCVRY] {hostname} Final data set:\n{json.dumps(mdns_data, indent=4, sort_keys=True)}"
        )

        # TODO could probably just put the call to cache update in the api_reachable method
        if update_cache:
            if 'hostname' in mdns_data[hostname]:
                del mdns_data[hostname]['hostname']
            cpi.remotes.data = cpi.remotes.update_local_cloud_file(remote_consoles=mdns_data)
            log.info(f'[MDNS DSCVRY] {hostname} Local Cache Updated after mdns discovery')

    def run(self):
        self.zc = Zeroconf()
        if not self.startup_logged:
            log.info(f"[MDNS DSCVRY] Discovering ConsolePis via mdns - Debug Enabled: {self.debug}")
            self.startup_logged = True
        return ServiceBrowser(self.zc, "_consolepi._tcp.local.", handlers=[self.on_service_state_change])  # NoQA pylint: disable=unused-variable


if __name__ == '__main__':
    program_start = int(time.time())
    if len(sys.argv) > 1:
        mdns = MDNS_Browser(show=True)
        RESTART_INTERVAL = 30  # when running in interactive mode reduce restart interval
        print("\nBrowsing services, press Ctrl-C to exit...\n")
    else:
        mdns = MDNS_Browser()

    try:
        while True:
            try:
                browser = mdns.run()
            except AttributeError:
                # hopefully this handles "Zeroconf object has no attribute '_handlers_lock'"
                log.warning('[MDNS BROWSE] caught _handlers_lock exception retrying in 5 sec')
                time.sleep(5)
                continue
            except Exception as e:
                # Catch any other errors, usually related to transient connectivity issues."
                log.warning(f'[MDNS BROWSE] caught {e.__class__.__name__} retrying in 5 sec.\nException:\n{e}')
                time.sleep(5)
                continue
            # re-init zeroconf browser every RESTART_INTERVAL seconds
            start = time.time()
            while time.time() < start + RESTART_INTERVAL:
                time.sleep(start + RESTART_INTERVAL - time.time())
            if mdns.zc is not None:
                mdns.zc.close()

            duration = f"{RESTART_INTERVAL / 60}m" if RESTART_INTERVAL > 60 else f"{RESTART_INTERVAL}s"
            mdns.discovered = []

    except KeyboardInterrupt:
        pass
    finally:
        if mdns.zc is not None:
            mdns.zc.close()



#!/etc/ConsolePi/venv/bin/python3

import sys

sys.path.insert(0, "/etc/ConsolePi/src/pypkg")
from consolepi import config, log, utils  # type: ignore # NoQA
from consolepi.consolepi import ConsolePi  # type: ignore # NoQA
from consolepi.gdrive import GoogleDrive  # type: ignore # NoQA


def main():
    cpi = ConsolePi()
    cloud_svc = config.cfg.get("cloud_svc", "error")
    local = cpi.local
    remotes = cpi.remotes
    cpiexec = cpi.cpiexec
    log.info("[CLOUD TRIGGER (IP)]: Cloud Update triggered by IP Update")
    CLOUD_CREDS_FILE = config.static.get("CLOUD_CREDS_FILE", "/etc/ConsolePi/cloud/gdrive/.credentials/credentials.json")
    if not utils.is_reachable("www.googleapis.com", 443):
        log.error(f"Not Updating {cloud_svc} due to connection failure")
        sys.exit(1)
    if not utils.valid_file(CLOUD_CREDS_FILE):
        log.error("Credentials file not found or invalid")
        sys.exit(1)

    # -- // Get details from Google Drive - once populated will skip \\ --
    if cloud_svc == "gdrive" and remotes.cloud is None:
        remotes.cloud = GoogleDrive(hostname=local.hostname)

    if cpiexec.wait_for_threads(thread_type="remotes") and (config.power and cpiexec.wait_for_threads(name="_toggle_refresh")):
        log.error("IP Change Cloud Update Trigger: TimeOut Waiting for Threads to Complete")

    remote_consoles = remotes.cloud.update_files(local.data)
    if remote_consoles and "Gdrive-Error:" in remote_consoles:
        log.error(remote_consoles)
    else:
        for r in remote_consoles:
            # -- Convert Any Remotes with old API schema to new API schema --
            if isinstance(remote_consoles[r].get("adapters", {}), list):
                remote_consoles[r]["adapters"] = remotes.convert_adapters(remote_consoles[r]["adapters"])
                log.warning(f"Adapter data for {r} retrieved from cloud in old API format... Converted")
        if len(remote_consoles) > 0:
            remotes.update_local_cloud_file(remote_consoles)

# TODO cmd line args when called from Network Dispatcher to bypass remotes (when updating on VPN down)
# Also can probably always bypass outlets
if __name__ == "__main__":
    main()



#!/etc/ConsolePi/venv/bin/python3

import time
import socket
from typing import Any, Dict, List, Union
from halo import Halo
from sys import stdin
from log_symbols import LogSymbols as log_sym  # Enum
from consolepi import utils, log, config, json  # type: ignore
from aiohttp import ClientSession
import asyncio
from aiohttp.client_exceptions import ContentTypeError, ClientConnectorError
# from pydantic import BaseModel
# from consolepi.gdrive import GoogleDrive  !!--> Import burried in refresh method to speed menu load times on older platforms

# Source = Literal["cloud", "mdns"]

# from .models import Remote


class Remotes:
    """Remotes Object Contains attributes for discovered remote ConsolePis

    bypass_cloud overrides the config value for cloud (gdrive sync)
        Used by mdns services which don't need cloud updates.
    """
    def __init__(self, local, cpiexec, bypass_cloud: bool = False):
        self.cpiexec = cpiexec
        self.pop_list = []
        self.old_api_log_sent = False
        self.log_sym_warn = log_sym.WARNING.value
        self.log_sym_error = log_sym.ERROR.value
        self.local = local
        self.connected: bool = False
        self.cache_update_pending = False
        self.spin = Halo(spinner="dots")
        self.running_spinners = []
        self.cloud = None  # Set in refresh method if reachable
        self.do_cloud = False if bypass_cloud is True else config.cfg.get("cloud", False)
        CLOUD_CREDS_FILE = config.static.get("CLOUD_CREDS_FILE")
        if not CLOUD_CREDS_FILE:
            self.no_creds_error()
        if self.do_cloud and config.cloud_svc == "gdrive":
            if utils.is_reachable("www.googleapis.com", 443, silent=False):
                self.local_only = False
                if not utils.valid_file(CLOUD_CREDS_FILE):
                    self.no_creds_error()
            else:
                log.warning(
                    f"failed to connect to {config.cloud_svc} - operating in local only mode",
                    show=True,
                )
                self.local_only = True
        self.data = asyncio.run(
            self.get_remote(
                data=config.remote_update()
            )  # re-get cloud.json to capture any updates via mdns
        )

    def no_creds_error(self):
        cloud_svc = config.cfg.get("cloud_svc", "UNDEFINED!")
        log.warning(
            f"Required {cloud_svc} credentials files are missing refer to GitHub for details"
        )
        log.warning(f"Disabling {cloud_svc} updates")
        log.show("Cloud Function Disabled by script - No Credentials Found")
        self.do_cloud = config.cfg["do_cloud"] = False

    # get remote consoles from local cache refresh function will check/update cloud file and update local cache
    async def get_remote(self, data: dict = None, rename: bool = False) -> Dict[str, Any]:
        spin = self.spin
        _get_remote_start = time.perf_counter()

        async def verify_remote(remotepi: str, data: dict, rename: bool) -> None:
            """sub to verify reachability and api data for remotes

            params:
            remotepi: The hostname currently being processed
            data: dict remote ConsolePi dict with hostname as key
            rename: bool set True to perform rename request (TODO not sure this is used.)
            """
            this = data[remotepi]
            if stdin.isatty():
                self.spin.stop()
                self.spin.start(f"verifying {remotepi}")
            self.running_spinners += [remotepi]
            res = await self.api_reachable(remotepi, this, rename=rename)
            _ = self.running_spinners.pop(self.running_spinners.index(remotepi))
            if stdin.isatty():  # restore spin text to any spinners that are still runnning
                self.spin.stop() if res.reachable else self.spin.fail(f'verifying {remotepi}')
                if self.running_spinners:
                    self.spin.start(f"verifying {self.running_spinners[-1]}")
                else:
                    if config.remotes:
                        self.spin.succeed(
                            "[GET REM] Querying Remotes via API to verify reachability and adapter data\n\t"
                            f"Found {len(config.remotes)} Remote ConsolePis in {time.perf_counter() - _get_remote_start:.2f}"
                        )
                    else:
                        self.spin.warn(
                            "[GET REM] Querying Remotes via API to verify reachability and adapter data\n\t"
                            "No Reachable Remote ConsolePis Discovered"
                        )

            this = res.data
            if res.update:
                self.cache_update_pending = True

            if not res.reachable:
                log.warning(
                    f"[GET REM] Found {remotepi} in Local Cloud Cache: UNREACHABLE"
                )
                this["fail_cnt"] = (this.get("fail_cnt", 0) + 1)
                self.pop_list += [remotepi]
                self.cache_update_pending = True
            else:
                self.connected = True
                if this.get("fail_cnt"):
                    this["fail_cnt"] = 0
                    self.cache_update_pending = True
                if res.update:
                    log.info(f"[GET REM] Updating Cache - Found {remotepi} in Local Cloud Cache, "
                             f"reachable via {this['rem_ip']}")

            data[remotepi] = this

        if data is None or len(data) == 0:
            data = config.remotes  # remotes from local cloud cache

        if not data:
            log.info("No Remotes found in Local Cache")
            data = {}  # convert None type to empy dict
        else:
            # if self is in the remote-data remove and warn user (can occur in rare scenarios i.e. hostname changes)
            if socket.gethostname() in data:
                del data[socket.gethostname()]
                log.show(
                    "Local cache included entry for self - do you have other ConsolePis using the same hostname?"
                )

            # Verify Remote ConsolePi details and reachability
            if stdin.isatty():
                spin.start(
                    "Querying Remotes via API to verify reachability and adapter data"
                )

            _ = await asyncio.gather(*[verify_remote(remotepi, data, rename) for remotepi in data])

        # update local cache if any ConsolePis found UnReachable
        if self.cache_update_pending:
            if self.pop_list:
                for remotepi in self.pop_list:
                    if (
                        data[remotepi]["fail_cnt"] >= 3
                    ):  # NoQA remove from local cache after 3 failures (cloud or mdns will repopulate if discovered)
                        removed = data.pop(remotepi)
                        log.warning(
                            "[GET REM] {} has been removed from Local Cache after {} failed attempts".format(
                                remotepi, removed["fail_cnt"]
                            ),
                            show=True,
                        )
                    else:
                        log.show(f"Cached Remote '{remotepi}' is unreachable")

            # update local cache file if rem_ip or adapter data changed
            # TODO check logic I think both this method and update_local_cloud_file are checking reachability
            #      During debug manually pop a device from data dict, then below command re-adds it think update_local_cloud_file
            #      is merging with config.remotes and checking reachability?
            data = self.update_local_cloud_file(data)
            self.pop_list = []
            self.cache_update_pending = False

        # TODO this is working, prob change adapters to a list of models
        # remotes = [Remote(**{"name": k, **data[k]}) for k in data]
        return data

    # Update with Data from ConsolePi.csv on Gdrive and local cache populated by mdns.  Update Gdrive with our data
    def refresh(self, bypass_cloud: bool = False):
        remote_consoles: Union[List[Dict[str, Any]], None] = None
        cpiexec = self.cpiexec
        local = self.local
        cloud_svc: str = config.cfg.get("cloud_svc", "error")

        # TODO refactor wait_for_threads to have an all key or accept a list
        with Halo(text="Waiting For threads to complete", spinner="dots1"):
            if cpiexec.wait_for_threads(thread_type="remotes") and (
                config.power and cpiexec.wait_for_threads(name="_toggle_refresh")
            ):
                log.show(
                    "Timeout Waiting for init or toggle threads to complete try again later or"
                    " investigate logs"
                )
                return

        # -- // Update/Refresh Local Data (Adapters/Interfaces) \\ --
        local.data = local.build_local_dict(refresh=True)
        log.debugv(f"Final Data set collected for {local.hostname}: {local.data}")

        # -- // Get details from Google Drive - once populated will skip \\ --
        if not bypass_cloud and self.do_cloud and not self.local_only:
            if cloud_svc == "gdrive" and self.cloud is None:
                # burried import until I find out why this import takes so @#%$@#% long.  Not imported until 1st refresh is called
                with Halo(text="Loading Google Drive Library", spinner="dots1"):
                    from consolepi.gdrive import GoogleDrive  # type: ignore
                self.cloud = GoogleDrive(hostname=local.hostname)
                log.info("[MENU REFRESH] Gdrive init")

            # Pass Local Data to update_sheet method get remotes found on sheet as return
            # update sheets function updates local_cloud_file
            _msg = "[MENU REFRESH] Updating to/from {}".format(cloud_svc)
            log.info(_msg)
            if stdin.isatty():
                self.spin.start(_msg)
            # -- // SYNC DATA WITH GDRIVE \\ --
            remote_consoles = self.cloud.update_files(
                local.data
            )  # local data refreshed above
            if remote_consoles and "Gdrive-Error:" not in remote_consoles:
                if stdin.isatty():
                    self.spin.succeed(f"{_msg}\n\tFound {len(remote_consoles)} Remotes via Gdrive Sync")
                    for r in remote_consoles:
                        # -- Convert Any Remotes with old API schema to new API schema --
                        if isinstance(remote_consoles[r].get("adapters", {}), list):
                            remote_consoles[r]["adapters"] = self.convert_adapters(
                                remote_consoles[r]["adapters"]
                            )
                            log.warning(
                                f"Adapter data for {r} retrieved from cloud in old API format... Converted"
                            )
            elif "Gdrive-Error:" in remote_consoles:
                if stdin.isatty():
                    self.spin.fail(
                        "{}\n\t{} {}".format(_msg, self.log_sym_error, remote_consoles)
                    )
                log.show(remote_consoles)  # display error returned from gdrive module
                remote_consoles = []
            else:
                if stdin.isatty():
                    self.spin.warn(_msg + "\n\tNo Remotes Found via Gdrive Sync")

            if len(remote_consoles) > 0:
                _msg = f"[MENU REFRESH] Updating Local Cache with data from {cloud_svc}"
                log.info(_msg)
                if stdin.isatty():
                    self.spin.start(_msg)
                self.update_local_cloud_file(remote_consoles)
                if stdin.isatty():
                    self.spin.succeed(_msg)  # no real error correction here
            else:
                log.warning(
                    f"[MENU REFRESH] No Remote ConsolePis found on {cloud_svc}",
                    show=True,
                )
        else:
            if self.do_cloud and not bypass_cloud:
                log.show(
                    f"Not Updating from {cloud_svc} due to connection failure\n"
                    "Close and re-launch menu if network access has been restored"
                )

        # Update Remote data with data from local_cloud cache / cloud
        self.data = asyncio.run(self.get_remote(data=remote_consoles))

    def update_local_cloud_file(
        self, remote_consoles=None, current_remotes=None, local_cloud_file=None
    ):
        """Update local cloud cache (cloud.json).

        Verifies the newly discovered data is more current than what we already know and updates the local cloud.json file if so
        The Menu uses cloud.json to populate remote menu items

        params:
            remote_consoles: The newly discovered data (from Gdrive or mdns)
            current_remotes: The current remote data fetched from the local cloud cache (cloud.json)
                - func will retrieve this if not provided
            local_cloud_file The path to the local cloud file (global var cloud.json)

        returns:
        dict: The resulting remote console dict representing the most recent data for each remote.
        """
        local_cloud_file = (
            config.static.get("LOCAL_CLOUD_FILE")
            if local_cloud_file is None
            else local_cloud_file
        )

        if len(remote_consoles) > 0:
            if current_remotes is None:
                current_remotes = self.data = config.remote_update()  # grabs the remote data from local cloud cache

        # update current_remotes dict with data passed to function
        if len(remote_consoles) > 0:
            if current_remotes is not None:
                for _ in current_remotes:
                    if _ not in remote_consoles:
                        if (
                            "fail_cnt" not in current_remotes[_]
                            or current_remotes[_]["fail_cnt"] < 2
                        ):
                            remote_consoles[_] = current_remotes[_]
                        elif (
                            remote_consoles.get(_)
                            and "fail_cnt" not in remote_consoles[_]
                            and "fail_cnt" in current_remotes[_]
                        ):
                            remote_consoles[_]["fail_cnt"] = current_remotes[_][
                                "fail_cnt"
                            ]
                    else:

                        # -- VERBOSE DEBUG --
                        log.debugv(
                            "[CACHE UPD] \n--{}-- \n    remote upd_time: {}\n    remote rem_ip: {}\n    remote source: {}\n    cache rem upd_time: {}\n    cache rem_ip: {}\n    cache source: {}\n".format(  # NoQA
                                _,
                                time.strftime(
                                    "%a %x %I:%M:%S %p %Z",
                                    time.localtime(remote_consoles[_]["upd_time"]),
                                )
                                if "upd_time" in remote_consoles[_]
                                else None,  # NoQA
                                remote_consoles[_]["rem_ip"]
                                if "rem_ip" in remote_consoles[_]
                                else None,
                                remote_consoles[_]["source"]
                                if "source" in remote_consoles[_]
                                else None,
                                time.strftime(
                                    "%a %x %I:%M:%S %p %Z",
                                    time.localtime(current_remotes[_]["upd_time"]),
                                )
                                if "upd_time" in current_remotes[_]
                                else None,  # NoQA
                                current_remotes[_]["rem_ip"]
                                if "rem_ip" in current_remotes[_]
                                else None,
                                current_remotes[_]["source"]
                                if "source" in current_remotes[_]
                                else None,
                            )
                        )
                        # -- END VERBOSE DEBUG --

                        # No Change Detected (data passed to function matches cache)
                        if "last_ip" in current_remotes[_]:
                            del current_remotes[_]["last_ip"]
                        if remote_consoles[_] == current_remotes[_]:
                            log.debug(
                                "[CACHE UPD] {} No Change in info detected".format(_)
                            )

                        # only factor in existing data if source is not mdns
                        elif (
                            "upd_time" in remote_consoles[_]
                            or "upd_time" in current_remotes[_]
                        ):
                            if (
                                "upd_time" in remote_consoles[_]
                                and "upd_time" in current_remotes[_]
                            ):
                                if (
                                    current_remotes[_]["upd_time"]
                                    > remote_consoles[_]["upd_time"]
                                ):
                                    remote_consoles[_] = current_remotes[_]
                                    log.info(
                                        f"[CACHE UPD] {_} Keeping existing data from {current_remotes[_].get('source', '')} "
                                        "based on more current update time"
                                    )
                                elif (
                                    remote_consoles[_]["upd_time"]
                                    > current_remotes[_]["upd_time"]
                                ):
                                    log.info(
                                        "[CACHE UPD] {} Updating data from {} "
                                        "based on more current update time".format(
                                            _, remote_consoles[_]["source"]
                                        )
                                    )
                                else:  # -- Update Times are equal --
                                    if (
                                        current_remotes[_].get("adapters")
                                        and remote_consoles[_].get("adapters")
                                        and current_remotes[_]["adapters"].keys()
                                        != remote_consoles[_]["adapters"].keys()
                                    ) or remote_consoles[_].get(
                                        "interfaces", {}
                                    ) != current_remotes[
                                        _
                                    ].get(
                                        "interfaces", {}
                                    ):
                                        log.warning(
                                            "[CACHE UPD] {} current cache update time and {} update time are equal"
                                            " but data appears to have changed. Updating".format(
                                                _, remote_consoles[_]["source"]
                                            )
                                        )
                            elif "upd_time" in current_remotes[_]:
                                remote_consoles[_] = current_remotes[_]
                                log.info(
                                    "[CACHE UPD] {} Keeping existing data based *existence* of update time "
                                    "which is lacking in this update from {}".format(
                                        _, remote_consoles[_]["source"]
                                    )
                                )

            for _ in range(0, 2):
                try:
                    with open(local_cloud_file, "w") as cloud_file:
                        cloud_file.write(
                            json.dumps(remote_consoles, indent=4, sort_keys=True)
                        )
                        utils.set_perm(
                            local_cloud_file
                        )  # a hack to deal with perms ~ consolepi-details del func
                        break
                except PermissionError:
                    utils.set_perm(local_cloud_file)

        else:
            log.warning(
                "[CACHE UPD] cache update called with no data passed, doing nothing"
            )

        return remote_consoles

    async def get_adapters_via_api(self, ip: str, port: int = 5000, rename: bool = False, log_host: str = None):
        """Send RestFul GET request to Remote ConsolePi to collect adapter info

        params:
        ip(str): ip address or FQDN of remote ConsolePi
        rename(bool): TODO
        log_host(str): friendly string for logging purposes "hostname(ip)"

        returns:
        adapter dict for remote if successful and adapters exist
        status_code 200 if successful but no adapters or Falsey or response status_code if an error occurred.
        22 if unable to reach API, but can reach port 22 (ssh)
        """
        if not log_host:
            log_host = ip
        url = f"http://{ip}:{port}/api/v1.0/adapters"
        if rename:
            url = f"{url}?refresh=true"

        log.debug(url)

        headers = {
            "Accept": "*/*",
            "Cache-Control": "no-cache",
            "Host": f"{ip}:{port}",
            "accept-encoding": "gzip, deflate",
            "Connection": "keep-alive",
            "cache-control": "no-cache",
        }

        ret = None
        try:
            _start = time.perf_counter()
            async with ClientSession() as client:
                resp = await client.request(
                    method="GET",
                    url=url,
                    headers=headers,
                    timeout=getattr(config.remote_timeout, log_host),
                )
                _elapsed = time.perf_counter() - _start
                if resp.ok:
                    try:
                        ret = await resp.json()
                        ret = ret["adapters"] if ret["adapters"] else resp.status
                        _msg = f"Adapters Successfully retrieved via API for Remote ConsolePi: {log_host}, elapsed {_elapsed:.2f}s"
                        log.info("[API RQST OUT] {}".format(_msg))
                        log.debugv(
                            "[API RQST OUT] Response: \n{}".format(
                                json.dumps(ret, indent=4, sort_keys=True)
                            )
                        )
                    except (json.decoder.JSONDecodeError, ContentTypeError):
                        log.error(f'[API RQST OUT] Puked on payload from {log_host} \n{await resp.text()}')
                        ret = resp.status
        except (asyncio.TimeoutError, ClientConnectorError):
            log.warning(f"[API RQST OUT] Remote ConsolePi: {log_host} TimeOut when querying via API - Unreachable.")
        except Exception as e:
            log.show(f'Exception: {e.__class__.__name__}, in remotes.get_adapters_via_api() check logs')
            log.exception(e)

        if not ret:  # hit an exception / or not reachable
            if utils.is_reachable(ip, port=22, silent=True):
                ret = 22  # indicates only available via ssh

        return ret

    async def api_reachable(self, remote_host: str, cache_data: dict, rename: bool = False):
        """Check Rechability & Fetch adapter data via API for remote ConsolePi

        params:
            remote_host:str, The hostname of the Remote ConsolePi
            cache_data:dict, The ConsolePi dictionary for the remote (from cache file)
            rename:bool, rename = True will do api call with refresh=True Query parameter
                which tells the api to first update connection data from ser2net as it likely
                changed as a result of remote rename operation.

        returns:
            APIReachableResponse object with the following attributes
                update: bool flag indicating if data was updated (I think)
                data: dict remote consolepi dict
                rechable: bool if the consolepi is reachable
        """

        class ApiReachableResponse:
            def __init__(self, update, data, reachable):
                self.update = update
                self.data = data
                self.reachable = reachable

        update = False
        local = self.local

        _iface_dict = cache_data["interfaces"]
        rem_ip_list = [
            _iface_dict[_iface].get("ip")
            for _iface in _iface_dict
            if not _iface.startswith("_") and _iface_dict[_iface].get("ip") not in local.ip_list
        ]

        # if inbound data includes rem_ip make sure to try that first
        for _ip in [cache_data.get("rem_ip"), cache_data.get("last_ip")]:
            if _ip:
                if _ip not in rem_ip_list:
                    rem_ip_list.insert(0, _ip)
                elif rem_ip_list.index(_ip) != 0:
                    rem_ip_list.remove(_ip)
                    rem_ip_list.insert(0, _ip)

        rem_ip = _adapters = None
        for _ip in rem_ip_list:
            log.debug(f"[API_REACHABLE] verifying {remote_host}")
            _adapters = await self.get_adapters_via_api(_ip, port=int(cache_data.get("api_port", 5000)), rename=rename, log_host=f"{remote_host}({_ip})")
            if _adapters:
                rem_ip = _ip  # Remote is reachable
                if not isinstance(_adapters, int):  # indicates status_code returned (error or no adapters found)
                    if isinstance(_adapters, list):  # indicates need for conversion from old api format
                        _adapters = self.convert_adapters(_adapters)
                        if not self.old_api_log_sent:
                            log.warning(
                                f"{remote_host} provided old api schema.  Recommend Upgrading to current."
                            )
                            self.old_api_log_sent = True
                    # Only compare config dict for each adapter as udev dict will generally be different due to time_since_init
                    if not cache_data.get("adapters") or {
                        a: {"config": _adapters[a].get("config", {})} for a in _adapters
                    } != {
                        a: {"config": cache_data["adapters"][a].get("config", {})}
                        for a in cache_data["adapters"]
                    }:
                        cache_data["adapters"] = _adapters
                        update = True  # --> Update if adapter dict is different
                    else:
                        cached_udev = [False for a in cache_data["adapters"] if 'udev' not in cache_data["adapters"][a]]
                        if False in cached_udev:
                            cache_data["adapters"] = _adapters
                            update = True  # --> Update if udev key not in existing data (udev not sent to cloud)
                elif _adapters == 200:
                    log.show(
                        f"Remote {remote_host} is reachable via {_ip},"
                        " but has no adapters attached\nit's still available in remote shell menu"
                    )
                elif _adapters == 22:
                    log.show(
                        f"Remote {remote_host}({_ip}) did not respond to API request,"
                        " but appears to be reachable via SSH\nit's available in remote shell menu"
                    )

                # remote was reachable update last_ip, even if returned bad status_code still reachable
                if not cache_data.get("last_ip", "") == _ip:
                    cache_data["last_ip"] = _ip
                    update = True  # --> Update if last_ip is different than currently reachable IP
                break

        if cache_data.get("rem_ip") != rem_ip:
            cache_data["rem_ip"] = rem_ip
            update = True  # --> Update if rem_ip didn't match (was previously unreachable)

        if not _adapters:
            reachable = False
            if isinstance(cache_data.get("adapters"), list):
                _adapters = cache_data.get("adapters")
                _adapters = {
                    _adapters[_adapters.index(d)]["dev"]: {
                        "config": {
                            k: _adapters[_adapters.index(d)][k]
                            for k in _adapters[_adapters.index(d)]
                        }
                    }
                    for d in _adapters
                }
                cache_data["adapters"] = _adapters
                _msg = (
                    f"{remote_host} Cached adapter data was in old format... Converted to new.\n"
                    f"\t\t{remote_host} Should be upgraded to the current version of ConsolePi."
                )
                log.warning(_msg, show=True)
                update = True  # --> Convert to new and Update if cache data was in old format
        else:
            reachable = True

        return ApiReachableResponse(update, cache_data, reachable)

    def convert_adapters(self, adapters):
        return {
            adapters[adapters.index(d)]["dev"]: {
                "config": {
                    k: adapters[adapters.index(d)][k]
                    for k in adapters[adapters.index(d)]
                }
            }
            for d in adapters
        }
